uniswap_v3_factory_abi = [
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "uint24",
                "name": "fee",
                "type": "uint24",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickSpacing",
                "type": "int24",
            },
        ],
        "name": "FeeAmountEnabled",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "oldOwner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "newOwner",
                "type": "address",
            },
        ],
        "name": "OwnerChanged",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "token0",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "token1",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "uint24",
                "name": "fee",
                "type": "uint24",
            },
            {
                "indexed": False,
                "internalType": "int24",
                "name": "tickSpacing",
                "type": "int24",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "pool",
                "type": "address",
            },
        ],
        "name": "PoolCreated",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "tokenA", "type": "address"},
            {"internalType": "address", "name": "tokenB", "type": "address"},
            {"internalType": "uint24", "name": "fee", "type": "uint24"},
        ],
        "name": "createPool",
        "outputs": [{"internalType": "address", "name": "pool", "type": "address"}],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint24", "name": "fee", "type": "uint24"},
            {"internalType": "int24", "name": "tickSpacing", "type": "int24"},
        ],
        "name": "enableFeeAmount",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint24", "name": "", "type": "uint24"}],
        "name": "feeAmountTickSpacing",
        "outputs": [{"internalType": "int24", "name": "", "type": "int24"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "address", "name": "", "type": "address"},
            {"internalType": "uint24", "name": "", "type": "uint24"},
        ],
        "name": "getPool",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "owner",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "parameters",
        "outputs": [
            {"internalType": "address", "name": "factory", "type": "address"},
            {"internalType": "address", "name": "token0", "type": "address"},
            {"internalType": "address", "name": "token1", "type": "address"},
            {"internalType": "uint24", "name": "fee", "type": "uint24"},
            {"internalType": "int24", "name": "tickSpacing", "type": "int24"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "_owner", "type": "address"}],
        "name": "setOwner",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]


uniswap_v3_router_abi = [
    {
        "inputs": [
            {"internalType": "address", "name": "_factoryV2", "type": "address"},
            {"internalType": "address", "name": "factoryV3", "type": "address"},
            {"internalType": "address", "name": "_positionManager", "type": "address"},
            {"internalType": "address", "name": "_WETH9", "type": "address"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "inputs": [],
        "name": "WETH9",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "approveMax",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "approveMaxMinusOne",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "approveZeroThenMax",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "address", "name": "token", "type": "address"}],
        "name": "approveZeroThenMaxMinusOne",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes", "name": "data", "type": "bytes"}],
        "name": "callPositionManager",
        "outputs": [{"internalType": "bytes", "name": "result", "type": "bytes"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes[]", "name": "paths", "type": "bytes[]"},
            {"internalType": "uint128[]", "name": "amounts", "type": "uint128[]"},
            {
                "internalType": "uint24",
                "name": "maximumTickDivergence",
                "type": "uint24",
            },
            {"internalType": "uint32", "name": "secondsAgo", "type": "uint32"},
        ],
        "name": "checkOracleSlippage",
        "outputs": [],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes", "name": "path", "type": "bytes"},
            {
                "internalType": "uint24",
                "name": "maximumTickDivergence",
                "type": "uint24",
            },
            {"internalType": "uint32", "name": "secondsAgo", "type": "uint32"},
        ],
        "name": "checkOracleSlippage",
        "outputs": [],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes", "name": "path", "type": "bytes"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountOutMinimum",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct IV3SwapRouter.ExactInputParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactInput",
        "outputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"}
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "tokenIn", "type": "address"},
                    {"internalType": "address", "name": "tokenOut", "type": "address"},
                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountOutMinimum",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint160",
                        "name": "sqrtPriceLimitX96",
                        "type": "uint160",
                    },
                ],
                "internalType": "struct IV3SwapRouter.ExactInputSingleParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactInputSingle",
        "outputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"}
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "bytes", "name": "path", "type": "bytes"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountInMaximum",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct IV3SwapRouter.ExactOutputParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactOutput",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "tokenIn", "type": "address"},
                    {"internalType": "address", "name": "tokenOut", "type": "address"},
                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                    {"internalType": "address", "name": "recipient", "type": "address"},
                    {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amountInMaximum",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint160",
                        "name": "sqrtPriceLimitX96",
                        "type": "uint160",
                    },
                ],
                "internalType": "struct IV3SwapRouter.ExactOutputSingleParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "exactOutputSingle",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "factory",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "factoryV2",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amount", "type": "uint256"},
        ],
        "name": "getApprovalType",
        "outputs": [
            {
                "internalType": "enum IApproveAndCall.ApprovalType",
                "name": "",
                "type": "uint8",
            }
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "token0", "type": "address"},
                    {"internalType": "address", "name": "token1", "type": "address"},
                    {"internalType": "uint256", "name": "tokenId", "type": "uint256"},
                    {
                        "internalType": "uint256",
                        "name": "amount0Min",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint256",
                        "name": "amount1Min",
                        "type": "uint256",
                    },
                ],
                "internalType": "struct IApproveAndCall.IncreaseLiquidityParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "increaseLiquidity",
        "outputs": [{"internalType": "bytes", "name": "result", "type": "bytes"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {"internalType": "address", "name": "token0", "type": "address"},
                    {"internalType": "address", "name": "token1", "type": "address"},
                    {"internalType": "uint24", "name": "fee", "type": "uint24"},
                    {"internalType": "int24", "name": "tickLower", "type": "int24"},
                    {"internalType": "int24", "name": "tickUpper", "type": "int24"},
                    {
                        "internalType": "uint256",
                        "name": "amount0Min",
                        "type": "uint256",
                    },
                    {
                        "internalType": "uint256",
                        "name": "amount1Min",
                        "type": "uint256",
                    },
                    {"internalType": "address", "name": "recipient", "type": "address"},
                ],
                "internalType": "struct IApproveAndCall.MintParams",
                "name": "params",
                "type": "tuple",
            }
        ],
        "name": "mint",
        "outputs": [{"internalType": "bytes", "name": "result", "type": "bytes"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "bytes32", "name": "previousBlockhash", "type": "bytes32"},
            {"internalType": "bytes[]", "name": "data", "type": "bytes[]"},
        ],
        "name": "multicall",
        "outputs": [{"internalType": "bytes[]", "name": "", "type": "bytes[]"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "bytes[]", "name": "data", "type": "bytes[]"},
        ],
        "name": "multicall",
        "outputs": [{"internalType": "bytes[]", "name": "", "type": "bytes[]"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes[]", "name": "data", "type": "bytes[]"}],
        "name": "multicall",
        "outputs": [{"internalType": "bytes[]", "name": "results", "type": "bytes[]"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "positionManager",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
        ],
        "name": "pull",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "refundETH",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermit",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "nonce", "type": "uint256"},
            {"internalType": "uint256", "name": "expiry", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitAllowed",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "nonce", "type": "uint256"},
            {"internalType": "uint256", "name": "expiry", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitAllowedIfNecessary",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "value", "type": "uint256"},
            {"internalType": "uint256", "name": "deadline", "type": "uint256"},
            {"internalType": "uint8", "name": "v", "type": "uint8"},
            {"internalType": "bytes32", "name": "r", "type": "bytes32"},
            {"internalType": "bytes32", "name": "s", "type": "bytes32"},
        ],
        "name": "selfPermitIfNecessary",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountIn", "type": "uint256"},
            {"internalType": "uint256", "name": "amountOutMin", "type": "uint256"},
            {"internalType": "address[]", "name": "path", "type": "address[]"},
            {"internalType": "address", "name": "to", "type": "address"},
        ],
        "name": "swapExactTokensForTokens",
        "outputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"}
        ],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountOut", "type": "uint256"},
            {"internalType": "uint256", "name": "amountInMax", "type": "uint256"},
            {"internalType": "address[]", "name": "path", "type": "address[]"},
            {"internalType": "address", "name": "to", "type": "address"},
        ],
        "name": "swapTokensForExactTokens",
        "outputs": [{"internalType": "uint256", "name": "amountIn", "type": "uint256"}],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
        ],
        "name": "sweepToken",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
        ],
        "name": "sweepToken",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "sweepTokenWithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "token", "type": "address"},
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "sweepTokenWithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "int256", "name": "amount0Delta", "type": "int256"},
            {"internalType": "int256", "name": "amount1Delta", "type": "int256"},
            {"internalType": "bytes", "name": "_data", "type": "bytes"},
        ],
        "name": "uniswapV3SwapCallback",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
        ],
        "name": "unwrapWETH9",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"}
        ],
        "name": "unwrapWETH9",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "unwrapWETH9WithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint256", "name": "amountMinimum", "type": "uint256"},
            {"internalType": "uint256", "name": "feeBips", "type": "uint256"},
            {"internalType": "address", "name": "feeRecipient", "type": "address"},
        ],
        "name": "unwrapWETH9WithFee",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "value", "type": "uint256"}],
        "name": "wrapETH",
        "outputs": [],
        "stateMutability": "payable",
        "type": "function",
    },
    {"stateMutability": "payable", "type": "receive"},
]

uniswap_v3_pool_abi = [
    {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickLower",
                "type": "int24",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickUpper",
                "type": "int24",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount",
                "type": "uint128",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256",
            },
        ],
        "name": "Burn",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "address",
                "name": "recipient",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickLower",
                "type": "int24",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickUpper",
                "type": "int24",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount0",
                "type": "uint128",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount1",
                "type": "uint128",
            },
        ],
        "name": "Collect",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "recipient",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount0",
                "type": "uint128",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount1",
                "type": "uint128",
            },
        ],
        "name": "CollectProtocol",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "recipient",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "paid0",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "paid1",
                "type": "uint256",
            },
        ],
        "name": "Flash",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint16",
                "name": "observationCardinalityNextOld",
                "type": "uint16",
            },
            {
                "indexed": False,
                "internalType": "uint16",
                "name": "observationCardinalityNextNew",
                "type": "uint16",
            },
        ],
        "name": "IncreaseObservationCardinalityNext",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint160",
                "name": "sqrtPriceX96",
                "type": "uint160",
            },
            {
                "indexed": False,
                "internalType": "int24",
                "name": "tick",
                "type": "int24",
            },
        ],
        "name": "Initialize",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "address",
                "name": "sender",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "owner",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickLower",
                "type": "int24",
            },
            {
                "indexed": True,
                "internalType": "int24",
                "name": "tickUpper",
                "type": "int24",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "amount",
                "type": "uint128",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount0",
                "type": "uint256",
            },
            {
                "indexed": False,
                "internalType": "uint256",
                "name": "amount1",
                "type": "uint256",
            },
        ],
        "name": "Mint",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "feeProtocol0Old",
                "type": "uint8",
            },
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "feeProtocol1Old",
                "type": "uint8",
            },
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "feeProtocol0New",
                "type": "uint8",
            },
            {
                "indexed": False,
                "internalType": "uint8",
                "name": "feeProtocol1New",
                "type": "uint8",
            },
        ],
        "name": "SetFeeProtocol",
        "type": "event",
    },
    {
        "anonymous": False,
        "inputs": [
            {
                "indexed": True,
                "internalType": "address",
                "name": "sender",
                "type": "address",
            },
            {
                "indexed": True,
                "internalType": "address",
                "name": "recipient",
                "type": "address",
            },
            {
                "indexed": False,
                "internalType": "int256",
                "name": "amount0",
                "type": "int256",
            },
            {
                "indexed": False,
                "internalType": "int256",
                "name": "amount1",
                "type": "int256",
            },
            {
                "indexed": False,
                "internalType": "uint160",
                "name": "sqrtPriceX96",
                "type": "uint160",
            },
            {
                "indexed": False,
                "internalType": "uint128",
                "name": "liquidity",
                "type": "uint128",
            },
            {
                "indexed": False,
                "internalType": "int24",
                "name": "tick",
                "type": "int24",
            },
        ],
        "name": "Swap",
        "type": "event",
    },
    {
        "inputs": [
            {"internalType": "int24", "name": "tickLower", "type": "int24"},
            {"internalType": "int24", "name": "tickUpper", "type": "int24"},
            {"internalType": "uint128", "name": "amount", "type": "uint128"},
        ],
        "name": "burn",
        "outputs": [
            {"internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"internalType": "uint256", "name": "amount1", "type": "uint256"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "int24", "name": "tickLower", "type": "int24"},
            {"internalType": "int24", "name": "tickUpper", "type": "int24"},
            {"internalType": "uint128", "name": "amount0Requested", "type": "uint128"},
            {"internalType": "uint128", "name": "amount1Requested", "type": "uint128"},
        ],
        "name": "collect",
        "outputs": [
            {"internalType": "uint128", "name": "amount0", "type": "uint128"},
            {"internalType": "uint128", "name": "amount1", "type": "uint128"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint128", "name": "amount0Requested", "type": "uint128"},
            {"internalType": "uint128", "name": "amount1Requested", "type": "uint128"},
        ],
        "name": "collectProtocol",
        "outputs": [
            {"internalType": "uint128", "name": "amount0", "type": "uint128"},
            {"internalType": "uint128", "name": "amount1", "type": "uint128"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "factory",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "fee",
        "outputs": [{"internalType": "uint24", "name": "", "type": "uint24"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "feeGrowthGlobal0X128",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "feeGrowthGlobal1X128",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"internalType": "uint256", "name": "amount1", "type": "uint256"},
            {"internalType": "bytes", "name": "data", "type": "bytes"},
        ],
        "name": "flash",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "internalType": "uint16",
                "name": "observationCardinalityNext",
                "type": "uint16",
            }
        ],
        "name": "increaseObservationCardinalityNext",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint160", "name": "sqrtPriceX96", "type": "uint160"}
        ],
        "name": "initialize",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "liquidity",
        "outputs": [{"internalType": "uint128", "name": "", "type": "uint128"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "maxLiquidityPerTick",
        "outputs": [{"internalType": "uint128", "name": "", "type": "uint128"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "int24", "name": "tickLower", "type": "int24"},
            {"internalType": "int24", "name": "tickUpper", "type": "int24"},
            {"internalType": "uint128", "name": "amount", "type": "uint128"},
            {"internalType": "bytes", "name": "data", "type": "bytes"},
        ],
        "name": "mint",
        "outputs": [
            {"internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"internalType": "uint256", "name": "amount1", "type": "uint256"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "name": "observations",
        "outputs": [
            {"internalType": "uint32", "name": "blockTimestamp", "type": "uint32"},
            {"internalType": "int56", "name": "tickCumulative", "type": "int56"},
            {
                "internalType": "uint160",
                "name": "secondsPerLiquidityCumulativeX128",
                "type": "uint160",
            },
            {"internalType": "bool", "name": "initialized", "type": "bool"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint32[]", "name": "secondsAgos", "type": "uint32[]"}
        ],
        "name": "observe",
        "outputs": [
            {"internalType": "int56[]", "name": "tickCumulatives", "type": "int56[]"},
            {
                "internalType": "uint160[]",
                "name": "secondsPerLiquidityCumulativeX128s",
                "type": "uint160[]",
            },
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "bytes32", "name": "", "type": "bytes32"}],
        "name": "positions",
        "outputs": [
            {"internalType": "uint128", "name": "liquidity", "type": "uint128"},
            {
                "internalType": "uint256",
                "name": "feeGrowthInside0LastX128",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "feeGrowthInside1LastX128",
                "type": "uint256",
            },
            {"internalType": "uint128", "name": "tokensOwed0", "type": "uint128"},
            {"internalType": "uint128", "name": "tokensOwed1", "type": "uint128"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "protocolFees",
        "outputs": [
            {"internalType": "uint128", "name": "token0", "type": "uint128"},
            {"internalType": "uint128", "name": "token1", "type": "uint128"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "uint8", "name": "feeProtocol0", "type": "uint8"},
            {"internalType": "uint8", "name": "feeProtocol1", "type": "uint8"},
        ],
        "name": "setFeeProtocol",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "slot0",
        "outputs": [
            {"internalType": "uint160", "name": "sqrtPriceX96", "type": "uint160"},
            {"internalType": "int24", "name": "tick", "type": "int24"},
            {"internalType": "uint16", "name": "observationIndex", "type": "uint16"},
            {
                "internalType": "uint16",
                "name": "observationCardinality",
                "type": "uint16",
            },
            {
                "internalType": "uint16",
                "name": "observationCardinalityNext",
                "type": "uint16",
            },
            {"internalType": "uint8", "name": "feeProtocol", "type": "uint8"},
            {"internalType": "bool", "name": "unlocked", "type": "bool"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "int24", "name": "tickLower", "type": "int24"},
            {"internalType": "int24", "name": "tickUpper", "type": "int24"},
        ],
        "name": "snapshotCumulativesInside",
        "outputs": [
            {"internalType": "int56", "name": "tickCumulativeInside", "type": "int56"},
            {
                "internalType": "uint160",
                "name": "secondsPerLiquidityInsideX128",
                "type": "uint160",
            },
            {"internalType": "uint32", "name": "secondsInside", "type": "uint32"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [
            {"internalType": "address", "name": "recipient", "type": "address"},
            {"internalType": "bool", "name": "zeroForOne", "type": "bool"},
            {"internalType": "int256", "name": "amountSpecified", "type": "int256"},
            {"internalType": "uint160", "name": "sqrtPriceLimitX96", "type": "uint160"},
            {"internalType": "bytes", "name": "data", "type": "bytes"},
        ],
        "name": "swap",
        "outputs": [
            {"internalType": "int256", "name": "amount0", "type": "int256"},
            {"internalType": "int256", "name": "amount1", "type": "int256"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "int16", "name": "", "type": "int16"}],
        "name": "tickBitmap",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "tickSpacing",
        "outputs": [{"internalType": "int24", "name": "", "type": "int24"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [{"internalType": "int24", "name": "", "type": "int24"}],
        "name": "ticks",
        "outputs": [
            {"internalType": "uint128", "name": "liquidityGross", "type": "uint128"},
            {"internalType": "int128", "name": "liquidityNet", "type": "int128"},
            {
                "internalType": "uint256",
                "name": "feeGrowthOutside0X128",
                "type": "uint256",
            },
            {
                "internalType": "uint256",
                "name": "feeGrowthOutside1X128",
                "type": "uint256",
            },
            {"internalType": "int56", "name": "tickCumulativeOutside", "type": "int56"},
            {
                "internalType": "uint160",
                "name": "secondsPerLiquidityOutsideX128",
                "type": "uint160",
            },
            {"internalType": "uint32", "name": "secondsOutside", "type": "uint32"},
            {"internalType": "bool", "name": "initialized", "type": "bool"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "token0",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "inputs": [],
        "name": "token1",
        "outputs": [{"internalType": "address", "name": "", "type": "address"}],
        "stateMutability": "view",
        "type": "function",
    },
]



from typing import TypedDict
from urllib.parse import urljoin

import requests
from eth_typing import ChecksumAddress, HexStr

from .. import SafeTx
from ..signatures import signature_split
from .base_api import SafeAPIException, SafeBaseAPI


class RelayEstimation(TypedDict):
    safeTxGas: int
    baseGas: int
    gasPrice: int
    lastUsedNonce: int
    gasToken: ChecksumAddress
    refundReceiver: ChecksumAddress


class RelaySentTransaction(TypedDict):
    safeTxHash: HexStr
    txHash: HexStr


class RelayServiceApi(SafeBaseAPI):
    URL_BY_NETWORK = {
        # Currently there's no official relayer operated
    }

    def send_transaction(
        self, safe_address: str, safe_tx: SafeTx
    ) -> RelaySentTransaction:
        url = urljoin(self.base_url, f"/api/v1/safes/{safe_address}/transactions/")
        signatures = []
        for i in range(len(safe_tx.signatures) // 65):
            v, r, s = signature_split(safe_tx.signatures, i)
            signatures.append(
                {
                    "v": v,
                    "r": r,
                    "s": s,
                }
            )

        data = {
            "to": safe_tx.to,
            "value": safe_tx.value,
            "data": safe_tx.data.hex() if safe_tx.data else None,
            "operation": safe_tx.operation,
            "gasToken": safe_tx.gas_token,
            "safeTxGas": safe_tx.safe_tx_gas,
            "dataGas": safe_tx.base_gas,
            "gasPrice": safe_tx.gas_price,
            "refundReceiver": safe_tx.refund_receiver,
            "nonce": safe_tx.safe_nonce,
            "signatures": signatures,
        }
        response = requests.post(url, json=data, timeout=self.request_timeout)
        if not response.ok:
            raise SafeAPIException(f"Error posting transaction: {response.content}")
        else:
            return RelaySentTransaction(response.json())

    def get_estimation(self, safe_address: str, safe_tx: SafeTx) -> RelayEstimation:
        """
        :param safe_address:
        :param safe_tx:
        :return: RelayEstimation
        """
        url = urljoin(
            self.base_url, f"/api/v2/safes/{safe_address}/transactions/estimate/"
        )
        data = {
            "to": safe_tx.to,
            "value": safe_tx.value,
            "data": safe_tx.data.hex() if safe_tx.data else None,
            "operation": safe_tx.operation,
            "gasToken": safe_tx.gas_token,
        }
        response = requests.post(url, json=data, timeout=self.request_timeout)
        if not response.ok:
            raise SafeAPIException(f"Error posting transaction: {response.content}")
        else:
            response_json = response.json()
            # Convert values to int
            for key in ("safeTxGas", "baseGas", "gasPrice"):
                response_json[key] = int(response_json[key])
            return RelayEstimation(response_json)



from functools import cached_property
from typing import Any, Dict, List, NoReturn, Optional, Tuple, Type

from eth_account import Account
from eth_typing import ChecksumAddress
from hexbytes import HexBytes
from packaging.version import Version
from web3.contract.contract import ContractFunction
from web3.exceptions import Web3Exception
from web3.types import BlockIdentifier, TxParams, Wei

from gnosis.eth import EthereumClient
from gnosis.eth.constants import NULL_ADDRESS
from gnosis.eth.contracts import get_safe_contract
from gnosis.eth.eip712 import eip712_encode
from gnosis.eth.ethereum_client import TxSpeed

from ..eth.utils import fast_keccak
from .exceptions import (
    CouldNotFinishInitialization,
    CouldNotPayGasWithEther,
    CouldNotPayGasWithToken,
    HashHasNotBeenApproved,
    InvalidContractSignatureLocation,
    InvalidInternalTx,
    InvalidMultisigTx,
    InvalidOwnerProvided,
    InvalidSignaturesProvided,
    MethodCanOnlyBeCalledFromThisContract,
    ModuleManagerException,
    NotEnoughSafeTransactionGas,
    OnlyOwnersCanApproveAHash,
    OwnerManagerException,
    SafeTransactionFailedWhenGasPriceAndSafeTxGasEmpty,
    SignatureNotProvidedByOwner,
    SignaturesDataTooShort,
    ThresholdNeedsToBeDefined,
)
from .safe_signature import SafeSignature
from .signatures import signature_to_bytes


class SafeTx:
    def __init__(
        self,
        ethereum_client: EthereumClient,
        safe_address: ChecksumAddress,
        to: Optional[ChecksumAddress],
        value: int,
        data: bytes,
        operation: int,
        safe_tx_gas: int,
        base_gas: int,
        gas_price: int,
        gas_token: Optional[ChecksumAddress],
        refund_receiver: Optional[ChecksumAddress],
        signatures: Optional[bytes] = None,
        safe_nonce: Optional[int] = None,
        safe_version: Optional[str] = None,
        chain_id: Optional[int] = None,
    ):
        """
        :param ethereum_client:
        :param safe_address:
        :param to:
        :param value:
        :param data:
        :param operation:
        :param safe_tx_gas:
        :param base_gas:
        :param gas_price:
        :param gas_token:
        :param refund_receiver:
        :param signatures:
        :param safe_nonce: Current nonce of the Safe. If not provided, it will be retrieved from network
        :param safe_version: Safe version 1.0.0 renamed `baseGas` to `dataGas`. Safe version 1.3.0 added `chainId` to
        the `domainSeparator`. If not provided, it will be retrieved from network
        :param chain_id: Ethereum network chain_id is used in hash calculation for Safes >= 1.3.0. If not provided,
        it will be retrieved from the provided ethereum_client
        """

        self.ethereum_client = ethereum_client
        self.safe_address = safe_address
        self.to = to or NULL_ADDRESS
        self.value = int(value)
        self.data = HexBytes(data) if data else b""
        self.operation = int(operation)
        self.safe_tx_gas = int(safe_tx_gas)
        self.base_gas = int(base_gas)
        self.gas_price = int(gas_price)
        self.gas_token = gas_token or NULL_ADDRESS
        self.refund_receiver = refund_receiver or NULL_ADDRESS
        self.signatures = signatures or b""
        self._safe_nonce = safe_nonce and int(safe_nonce)
        self._safe_version = safe_version
        self._chain_id = chain_id and int(chain_id)

        self.tx: Optional[TxParams] = None  # If executed, `tx` is set
        self.tx_hash: Optional[bytes] = None  # If executed, `tx_hash` is set

    def __str__(self):
        return (
            f"SafeTx - safe={self.safe_address} - to={self.to} - value={self.value} - data={self.data.hex()} - "
            f"operation={self.operation} - safe-tx-gas={self.safe_tx_gas} - base-gas={self.base_gas} - "
            f"gas-price={self.gas_price} - gas-token={self.gas_token} - refund-receiver={self.refund_receiver} - "
            f"signers = {self.signers}"
        )

    @property
    def w3(self):
        return self.ethereum_client.w3

    @cached_property
    def contract(self):
        return get_safe_contract(self.w3, address=self.safe_address)

    @cached_property
    def chain_id(self) -> int:
        if self._chain_id is not None:
            return self._chain_id
        else:
            return self.ethereum_client.get_chain_id()

    @cached_property
    def safe_nonce(self) -> int:
        if self._safe_nonce is not None:
            return self._safe_nonce
        else:
            return self.contract.functions.nonce().call()

    @cached_property
    def safe_version(self) -> str:
        if self._safe_version is not None:
            return self._safe_version
        else:
            return self.contract.functions.VERSION().call()

    @property
    def eip712_structured_data(self) -> Dict[str, Any]:
        safe_version = Version(self.safe_version)

        # Safes >= 1.0.0 Renamed `baseGas` to `dataGas`
        base_gas_key = "baseGas" if safe_version >= Version("1.0.0") else "dataGas"

        types = {
            "EIP712Domain": [{"name": "verifyingContract", "type": "address"}],
            "SafeTx": [
                {"name": "to", "type": "address"},
                {"name": "value", "type": "uint256"},
                {"name": "data", "type": "bytes"},
                {"name": "operation", "type": "uint8"},
                {"name": "safeTxGas", "type": "uint256"},
                {"name": base_gas_key, "type": "uint256"},
                {"name": "gasPrice", "type": "uint256"},
                {"name": "gasToken", "type": "address"},
                {"name": "refundReceiver", "type": "address"},
                {"name": "nonce", "type": "uint256"},
            ],
        }
        message = {
            "to": self.to,
            "value": self.value,
            "data": self.data,
            "operation": self.operation,
            "safeTxGas": self.safe_tx_gas,
            base_gas_key: self.base_gas,
            "dataGas": self.base_gas,
            "gasPrice": self.gas_price,
            "gasToken": self.gas_token,
            "refundReceiver": self.refund_receiver,
            "nonce": self.safe_nonce,
        }

        payload = {
            "types": types,
            "primaryType": "SafeTx",
            "domain": {"verifyingContract": self.safe_address},
            "message": message,
        }

        # Enable chainId from v1.3.0 onwards
        if safe_version >= Version("1.3.0"):
            payload["domain"]["chainId"] = self.chain_id
            types["EIP712Domain"].insert(0, {"name": "chainId", "type": "uint256"})

        return payload

    @property
    def safe_tx_hash_preimage(self) -> HexBytes:
        return HexBytes(b"".join(eip712_encode(self.eip712_structured_data)))

    @property
    def safe_tx_hash(self) -> HexBytes:
        return HexBytes(fast_keccak(self.safe_tx_hash_preimage))

    @property
    def signers(self) -> List[str]:
        if not self.signatures:
            return []
        else:
            return [
                safe_signature.owner
                for safe_signature in SafeSignature.parse_signature(
                    self.signatures, self.safe_tx_hash
                )
            ]

    @property
    def sorted_signers(self):
        return sorted(self.signers, key=lambda x: int(x, 16))

    @property
    def w3_tx(self) -> ContractFunction:
        """
        :return: Web3 contract tx prepared for `call`, `transact` or `build_transaction`
        """
        return self.contract.functions.execTransaction(
            self.to,
            self.value,
            self.data,
            self.operation,
            self.safe_tx_gas,
            self.base_gas,
            self.gas_price,
            self.gas_token,
            self.refund_receiver,
            self.signatures,
        )

    def _raise_safe_vm_exception(self, message: str) -> NoReturn:
        error_with_exception: Dict[str, Type[InvalidMultisigTx]] = {
            # https://github.com/safe-global/safe-contracts/blob/v1.3.0/docs/error_codes.md
            "GS000": CouldNotFinishInitialization,
            "GS001": ThresholdNeedsToBeDefined,
            "Could not pay gas costs with ether": CouldNotPayGasWithEther,
            "GS011": CouldNotPayGasWithEther,
            "Could not pay gas costs with token": CouldNotPayGasWithToken,
            "GS012": CouldNotPayGasWithToken,
            "GS013": SafeTransactionFailedWhenGasPriceAndSafeTxGasEmpty,
            "Hash has not been approved": HashHasNotBeenApproved,
            "Hash not approved": HashHasNotBeenApproved,
            "GS025": HashHasNotBeenApproved,
            "Invalid contract signature location: data not complete": InvalidContractSignatureLocation,
            "GS023": InvalidContractSignatureLocation,
            "Invalid contract signature location: inside static part": InvalidContractSignatureLocation,
            "GS021": InvalidContractSignatureLocation,
            "Invalid contract signature location: length not present": InvalidContractSignatureLocation,
            "GS022": InvalidContractSignatureLocation,
            "Invalid contract signature provided": InvalidContractSignatureLocation,
            "GS024": InvalidContractSignatureLocation,
            "Invalid owner provided": InvalidOwnerProvided,
            "Invalid owner address provided": InvalidOwnerProvided,
            "GS026": InvalidOwnerProvided,
            "Invalid signatures provided": InvalidSignaturesProvided,
            "Not enough gas to execute safe transaction": NotEnoughSafeTransactionGas,
            "GS010": NotEnoughSafeTransactionGas,
            "Only owners can approve a hash": OnlyOwnersCanApproveAHash,
            "GS030": OnlyOwnersCanApproveAHash,
            "GS031": MethodCanOnlyBeCalledFromThisContract,
            "Signature not provided by owner": SignatureNotProvidedByOwner,
            "Signatures data too short": SignaturesDataTooShort,
            "GS020": SignaturesDataTooShort,
            # ModuleManager
            "GS100": ModuleManagerException,
            "Invalid module address provided": ModuleManagerException,
            "GS101": ModuleManagerException,
            "GS102": ModuleManagerException,
            "Invalid prevModule, module pair provided": ModuleManagerException,
            "GS103": ModuleManagerException,
            "Method can only be called from an enabled module": ModuleManagerException,
            "GS104": ModuleManagerException,
            "Module has already been added": ModuleManagerException,
            # OwnerManager
            "Address is already an owner": OwnerManagerException,
            "GS200": OwnerManagerException,  # Owners have already been setup
            "GS201": OwnerManagerException,  # Threshold cannot exceed owner count
            "GS202": OwnerManagerException,  # Invalid owner address provided
            "GS203": OwnerManagerException,  # Invalid ower address provided
            "GS204": OwnerManagerException,  # Address is already an owner
            "GS205": OwnerManagerException,  # Invalid prevOwner, owner pair provided
            "Invalid prevOwner, owner pair provided": OwnerManagerException,
            "New owner count needs to be larger than new threshold": OwnerManagerException,
            "Threshold cannot exceed owner count": OwnerManagerException,
            "Threshold needs to be greater than 0": OwnerManagerException,
        }

        for reason, custom_exception in error_with_exception.items():
            if reason in message:
                raise custom_exception(message)
        raise InvalidMultisigTx(message)

    def call(
        self,
        tx_sender_address: Optional[str] = None,
        tx_gas: Optional[int] = None,
        block_identifier: Optional[BlockIdentifier] = "latest",
    ) -> int:
        """
        :param tx_sender_address:
        :param tx_gas: Force a gas limit
        :param block_identifier:
        :return: `1` if everything ok
        """
        parameters: Dict[str, Any] = {
            "from": tx_sender_address if tx_sender_address else self.safe_address
        }

        if tx_gas:
            parameters["gas"] = tx_gas
        try:
            success = self.w3_tx.call(parameters, block_identifier=block_identifier)

            if not success:
                raise InvalidInternalTx(
                    "Success bit is %d, should be equal to 1" % success
                )
            return success
        except (Web3Exception, ValueError) as exc:
            # e.g. web3.exceptions.ContractLogicError: execution reverted: Invalid owner provided
            return self._raise_safe_vm_exception(str(exc))
        except ValueError as exc:  # Parity
            """
            Parity throws a ValueError, e.g.
            {'code': -32015,
             'message': 'VM execution error.',
             'data': 'Reverted 0x08c379a0000000000000000000000000000000000000000000000000000000000000020000000000000000
                      000000000000000000000000000000000000000000000001b496e76616c6964207369676e6174757265732070726f7669
                      6465640000000000'
            }
            """
            error_dict = exc.args[0]
            data = error_dict.get("data")
            if data and isinstance(data, str) and "Reverted " in data:
                # Parity
                result = HexBytes(data.replace("Reverted ", ""))
                return self._raise_safe_vm_exception(str(result))
            else:
                raise exc

    def recommended_gas(self) -> Wei:
        """
        :return: Recommended gas to use on the ethereum_tx
        """
        return Wei(self.base_gas + self.safe_tx_gas + 75000)

    def execute(
        self,
        tx_sender_private_key: str,
        tx_gas: Optional[int] = None,
        tx_gas_price: Optional[int] = None,
        tx_nonce: Optional[int] = None,
        block_identifier: Optional[BlockIdentifier] = "latest",
        eip1559_speed: Optional[TxSpeed] = None,
    ) -> Tuple[HexBytes, TxParams]:
        """
        Send multisig tx to the Safe

        :param tx_sender_private_key: Sender private key
        :param tx_gas: Gas for the external tx. If not, `(safe_tx_gas + base_gas) * 2` will be used
        :param tx_gas_price: Gas price of the external tx. If not, `gas_price` will be used
        :param tx_nonce: Force nonce for `tx_sender`
        :param block_identifier: `latest` or `pending`
        :param eip1559_speed: If provided, use EIP1559 transaction
        :return: Tuple(tx_hash, tx)
        :raises: InvalidMultisigTx: If user tx cannot go through the Safe
        """

        sender_account = Account.from_key(tx_sender_private_key)
        if eip1559_speed and self.ethereum_client.is_eip1559_supported():
            tx_parameters = self.ethereum_client.set_eip1559_fees(
                {
                    "from": sender_account.address,
                },
                tx_speed=eip1559_speed,
            )
        else:
            tx_parameters = {
                "from": sender_account.address,
                "gasPrice": tx_gas_price or self.w3.eth.gas_price,
            }

        if tx_gas:
            tx_parameters["gas"] = tx_gas
        if tx_nonce is not None:
            tx_parameters["nonce"] = tx_nonce

        self.tx: TxParams = self.w3_tx.build_transaction(tx_parameters)
        self.tx["gas"] = Wei(
            tx_gas or (max(self.tx["gas"] + 75000, self.recommended_gas()))
        )

        self.tx_hash = self.ethereum_client.send_unsigned_transaction(
            self.tx,
            private_key=sender_account.key,
            retry=False if tx_nonce is not None else True,
            block_identifier=block_identifier,
        )

        # Set signatures empty after executing the tx. `Nonce` is increased even if it fails,
        # so signatures are not valid anymore
        self.signatures = b""
        return self.tx_hash, self.tx

    def sign(self, private_key: str) -> bytes:
        """
        {bytes32 r}{bytes32 s}{uint8 v}
        :param private_key:
        :return: Signature
        """
        account = Account.from_key(private_key)
        signature_dict = account.signHash(self.safe_tx_hash)
        signature = signature_to_bytes(
            signature_dict["v"], signature_dict["r"], signature_dict["s"]
        )

        # Insert signature sorted
        if account.address not in self.signers:
            new_owners = self.signers + [account.address]
            new_owner_pos = sorted(new_owners, key=lambda x: int(x, 16)).index(
                account.address
            )
            self.signatures = (
                self.signatures[: 65 * new_owner_pos]
                + signature
                + self.signatures[65 * new_owner_pos :]
            )
        return signature

    def unsign(self, address: str) -> bool:
        for pos, signer in enumerate(self.signers):
            if signer == address:
                self.signatures = self.signatures.replace(
                    self.signatures[pos * 65 : pos * 65 + 65], b""
                )
                return True
        return False



import logging
from typing import Any, Dict, List, Optional, Tuple, Union

from eth_account.signers.local import LocalAccount
from eth_typing import ChecksumAddress, HexStr
from hexbytes import HexBytes

from gnosis.eth import EthereumNetwork
from gnosis.eth.utils import fast_keccak_text
from gnosis.safe import SafeTx

from ..base_api import SafeAPIException, SafeBaseAPI
from .transaction_service_messages import get_delegate_message
from .transaction_service_tx import TransactionServiceTx

logger = logging.getLogger(__name__)


class TransactionServiceApi(SafeBaseAPI):
    URL_BY_NETWORK = {
        EthereumNetwork.ARBITRUM_ONE: "https://safe-transaction-arbitrum.safe.global",
        EthereumNetwork.AURORA_MAINNET: "https://safe-transaction-aurora.safe.global",
        EthereumNetwork.AVALANCHE_C_CHAIN: "https://safe-transaction-avalanche.safe.global",
        EthereumNetwork.BASE_GOERLI_TESTNET: "https://safe-transaction-base-testnet.safe.global",
        EthereumNetwork.BASE: "https://safe-transaction-base.safe.global",
        EthereumNetwork.BNB_SMART_CHAIN_MAINNET: "https://safe-transaction-bsc.safe.global",
        EthereumNetwork.CELO_MAINNET: "https://safe-transaction-celo.safe.global",
        EthereumNetwork.GNOSIS: "https://safe-transaction-gnosis-chain.safe.global",
        EthereumNetwork.GOERLI: "https://safe-transaction-goerli.safe.global",
        EthereumNetwork.MAINNET: "https://safe-transaction-mainnet.safe.global",
        EthereumNetwork.OPTIMISM: "https://safe-transaction-optimism.safe.global",
        EthereumNetwork.POLYGON: "https://safe-transaction-polygon.safe.global",
        EthereumNetwork.POLYGON_ZKEVM: "https://safe-transaction-zkevm.safe.global",
        EthereumNetwork.SEPOLIA: "https://safe-transaction-sepolia.safe.global",
        EthereumNetwork.ZKSYNC_MAINNET: "https://safe-transaction-zksync.safe.global",
    }

    @classmethod
    def create_delegate_message_hash(cls, delegate_address: ChecksumAddress) -> str:
        return fast_keccak_text(get_delegate_message(delegate_address))

    @classmethod
    def data_decoded_to_text(cls, data_decoded: Dict[str, Any]) -> Optional[str]:
        """
        Decoded data decoded to text
        :param data_decoded:
        :return:
        """
        if not data_decoded:
            return None

        method = data_decoded["method"]
        parameters = data_decoded.get("parameters", [])
        text = ""
        for (
            parameter
        ) in parameters:  # Multisend or executeTransaction from another Safe
            if "decodedValue" in parameter:
                text += (
                    method
                    + ":\n - "
                    + "\n - ".join(
                        [
                            cls.data_decoded_to_text(
                                decoded_value.get("decodedData", {})
                            )
                            for decoded_value in parameter.get("decodedValue", {})
                        ]
                    )
                    + "\n"
                )
        if text:
            return text.strip()
        else:
            return (
                method
                + ": "
                + ",".join([str(parameter["value"]) for parameter in parameters])
            )

    @classmethod
    def parse_signatures(cls, raw_tx: Dict[str, Any]) -> Optional[bytes]:
        """
        Parse signatures in `confirmations` list to build a valid signature (owners must be sorted lexicographically)

        :param raw_tx:
        :return: Valid signature with signatures sorted lexicographically
        """
        if raw_tx["signatures"]:
            # Tx was executed and signatures field is populated
            return raw_tx["signatures"]
        elif raw_tx["confirmations"]:
            # Parse offchain transactions
            return b"".join(
                [
                    HexBytes(confirmation["signature"])
                    for confirmation in sorted(
                        raw_tx["confirmations"], key=lambda x: int(x["owner"], 16)
                    )
                    if confirmation["signatureType"] == "EOA"
                ]
            )

    def get_balances(self, safe_address: str) -> List[Dict[str, Any]]:
        """

        :param safe_address:
        :return: a list of balances for provided Safe
        """
        response = self._get_request(f"/api/v1/safes/{safe_address}/balances/")
        if not response.ok:
            raise SafeAPIException(f"Cannot get balances: {response.content}")
        return response.json()

    def get_safe_transaction(
        self, safe_tx_hash: Union[bytes, HexStr]
    ) -> Tuple[TransactionServiceTx, Optional[HexBytes]]:
        """
        :param safe_tx_hash:
        :return: SafeTx and `tx-hash` if transaction was executed
        """
        safe_tx_hash = HexBytes(safe_tx_hash).hex()
        response = self._get_request(f"/api/v1/multisig-transactions/{safe_tx_hash}/")
        if not response.ok:
            raise SafeAPIException(
                f"Cannot get transaction with safe-tx-hash={safe_tx_hash}: {response.content}"
            )

        result = response.json()
        signatures = self.parse_signatures(result)
        if not self.ethereum_client:
            logger.warning(
                "EthereumClient should be defined to get a executable SafeTx"
            )
        safe_tx = TransactionServiceTx(
            result["proposer"],
            self.ethereum_client,
            result["safe"],
            result["to"],
            int(result["value"]),
            HexBytes(result["data"]) if result["data"] else b"",
            int(result["operation"]),
            int(result["safeTxGas"]),
            int(result["baseGas"]),
            int(result["gasPrice"]),
            result["gasToken"],
            result["refundReceiver"],
            signatures=signatures if signatures else b"",
            safe_nonce=int(result["nonce"]),
            chain_id=self.network.value,
        )
        tx_hash = (
            HexBytes(result["transactionHash"]) if result["transactionHash"] else None
        )
        if tx_hash:
            safe_tx.tx_hash = tx_hash
        return safe_tx, tx_hash

    def get_transactions(self, safe_address: ChecksumAddress) -> List[Dict[str, Any]]:
        """

        :param safe_address:
        :return: a list of transactions for provided Safe
        """
        response = self._get_request(
            f"/api/v1/safes/{safe_address}/multisig-transactions/"
        )
        if not response.ok:
            raise SafeAPIException(f"Cannot get transactions: {response.content}")
        return response.json().get("results", [])

    def get_delegates(self, safe_address: ChecksumAddress) -> List[Dict[str, Any]]:
        """

        :param safe_address:
        :return: a list of delegates for provided Safe
        """
        response = self._get_request(f"/api/v1/delegates/?safe={safe_address}")
        if not response.ok:
            raise SafeAPIException(f"Cannot get delegates: {response.content}")
        return response.json().get("results", [])

    def get_safes_for_owner(
        self, owner_address: ChecksumAddress
    ) -> List[ChecksumAddress]:
        """

        :param owner_address:
        :return: a List of Safe addresses which the owner_address is an owner
        """
        response = self._get_request(f"/api/v1/owners/{owner_address}/safes/")
        if not response.ok:
            raise SafeAPIException(f"Cannot get delegates: {response.content}")
        return response.json().get("safes", [])

    def post_signatures(self, safe_tx_hash: bytes, signatures: bytes) -> bool:
        """
        Create a new confirmation with provided signature for the given safe_tx_hash
        :param safe_tx_hash:
        :param signatures:
        :return: True if new confirmation was created
        """
        safe_tx_hash = HexBytes(safe_tx_hash).hex()
        response = self._post_request(
            f"/api/v1/multisig-transactions/{safe_tx_hash}/confirmations/",
            payload={"signature": HexBytes(signatures).hex()},
        )
        if not response.ok:
            raise SafeAPIException(
                f"Cannot post signatures for tx with safe-tx-hash={safe_tx_hash}: {response.content}"
            )
        return True

    def add_delegate(
        self,
        safe_address: ChecksumAddress,
        delegate_address: ChecksumAddress,
        label: str,
        signer_account: LocalAccount,
    ) -> bool:
        hash_to_sign = self.create_delegate_message_hash(delegate_address)
        signature = signer_account.signHash(hash_to_sign)
        add_payload = {
            "safe": safe_address,
            "delegate": delegate_address,
            "signature": signature.signature.hex(),
            "label": label,
        }
        response = self._post_request(
            f"/api/v1/safes/{safe_address}/delegates/", add_payload
        )
        if not response.ok:
            raise SafeAPIException(f"Cannot add delegate: {response.content}")
        return True

    def remove_delegate(
        self,
        safe_address: ChecksumAddress,
        delegate_address: ChecksumAddress,
        signer_account: LocalAccount,
    ) -> bool:
        hash_to_sign = self.create_delegate_message_hash(delegate_address)
        signature = signer_account.signHash(hash_to_sign)
        remove_payload = {"signature": signature.signature.hex()}
        response = self._delete_request(
            f"/api/v1/safes/{safe_address}/delegates/{delegate_address}/",
            remove_payload,
        )
        if not response.ok:
            raise SafeAPIException(f"Cannot remove delegate: {response.content}")
        return True

    def post_transaction(self, safe_tx: SafeTx) -> bool:
        random_sender = "0x0000000000000000000000000000000000000002"
        sender = safe_tx.sorted_signers[0] if safe_tx.sorted_signers else random_sender
        data = {
            "to": safe_tx.to,
            "value": safe_tx.value,
            "data": safe_tx.data.hex() if safe_tx.data else None,
            "operation": safe_tx.operation,
            "gasToken": safe_tx.gas_token,
            "safeTxGas": safe_tx.safe_tx_gas,
            "baseGas": safe_tx.base_gas,
            "gasPrice": safe_tx.gas_price,
            "refundReceiver": safe_tx.refund_receiver,
            "nonce": safe_tx.safe_nonce,
            "contractTransactionHash": safe_tx.safe_tx_hash.hex(),
            "sender": sender,
            "signature": safe_tx.signatures.hex() if safe_tx.signatures else None,
            "origin": "Safe-CLI",
        }
        response = self._post_request(
            f"/api/v1/safes/{safe_tx.safe_address}/multisig-transactions/", data
        )
        if not response.ok:
            raise SafeAPIException(f"Error posting transaction: {response.content}")
        return True

    def delete_transaction(self, safe_tx_hash: str, signature: str) -> bool:
        """

        :param safe_tx_hash: hash of eip712 see in transaction_service_messages.py generate_remove_transaction_message function
        :param signature: signature of safe_tx_hash by transaction proposer
        :return:
        """
        payload = {"safeTxHash": safe_tx_hash, "signature": signature}
        response = self._delete_request(
            f"/api/v1/multisig-transactions/{safe_tx_hash}/", payload
        )
        if not response.ok:
            raise SafeAPIException(f"Error deleting transaction: {response.content}")
        return True

    def post_message(
        self,
        safe_address: ChecksumAddress,
        message: Union[str, Dict],
        signature: bytes,
        safe_app_id: Optional[int] = 0,
    ) -> bool:
        """
        Create safe message on transaction service for provided Safe address

        :param safe_address:
        :param message: If str it will be encoded using EIP191, and if it's a dictionary it will be encoded using EIP712
        :param signature:
        :return:
        """
        payload = {
            "message": message,
            "safeAppId": safe_app_id,
            "signature": HexBytes(signature).hex(),
        }
        response = self._post_request(
            f"/api/v1/safes/{safe_address}/messages/", payload
        )
        if not response.ok:
            raise SafeAPIException(f"Error posting message: {response.content}")
        return True

    def get_message(self, safe_message_hash: bytes) -> Dict[str, Any]:
        """

        :param safe_message_hash:
        :return: Safe message for provided Safe message hash
        """
        response = self._get_request(
            f"/api/v1/messages/{HexBytes(safe_message_hash).hex()}/"
        )
        if not response.ok:
            raise SafeAPIException(f"Cannot get messages: {response.content}")
        return response.json()

    def get_messages(self, safe_address: ChecksumAddress) -> List[Dict[str, Any]]:
        """

        :param safe_address:
        :return: list of messages for provided Safe address
        """
        response = self._get_request(f"/api/v1/safes/{safe_address}/messages/")
        if not response.ok:
            raise SafeAPIException(f"Cannot get messages: {response.content}")
        return response.json().get("results", [])

    def post_message_signature(
        self, safe_message_hash: bytes, signature: bytes
    ) -> bool:
        """
        Add a new confirmation for provided Safe message hash

        :param safe_message_hash:
        :param signature:
        :return:
        """
        payload = {"signature": HexBytes(signature).hex()}
        response = self._post_request(
            f"/api/v1/messages/{HexBytes(safe_message_hash).hex()}/signatures/", payload
        )
        if not response.ok:
            raise SafeAPIException(
                f"Error posting message signature: {response.content}"
            )
        return True



import logging

from django.test import TestCase

from eth_account import Account

from gnosis.eth.contracts import get_safe_contract, get_safe_V1_0_0_contract

from ..safe_create2_tx import SafeCreate2TxBuilder
from .safe_test_case import SafeTestCaseMixin
from .utils import generate_salt_nonce

logger = logging.getLogger(__name__)

LOG_TITLE_WIDTH = 100


class TestSafeCreation2Tx(SafeTestCaseMixin, TestCase):
    def test_safe_create2_tx_builder(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(
            w3=w3,
            master_copy_address=self.safe_contract.address,
            proxy_factory_address=self.proxy_factory_contract.address,
        ).build(
            owners=owners,
            threshold=threshold,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
        )

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx(
            {
                "to": safe_creation_tx.safe_address,
                "value": safe_creation_tx.payment,
            },
            funder_account,
        )

        funder_balance = self.ethereum_client.get_balance(funder_account.address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            funder_account,
            self.safe_contract.address,
            safe_creation_tx.safe_setup_data,
            salt_nonce,
            safe_creation_tx.gas,
            safe_creation_tx.gas_price,
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt["status"], 1)

        # Funder balance must be bigger after a Safe deployment, as Safe deployment is a little overpriced
        self.assertGreater(
            self.ethereum_client.get_balance(funder_account.address), funder_balance
        )
        logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
            tx_receipt
        )
        log = logs[0]
        self.assertIsNone(tx_receipt["contractAddress"])
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
        )

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getThreshold().call(), threshold
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getOwners().call(), owners
        )
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_v_1_0_0(self):
        w3 = self.w3
        tx_hash = (
            get_safe_V1_0_0_contract(self.w3)
            .constructor()
            .transact({"from": self.ethereum_test_account.address})
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        master_copy = tx_receipt["contractAddress"]

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(
            w3=w3,
            master_copy_address=master_copy,
            proxy_factory_address=self.proxy_factory_contract.address,
        ).build(
            owners=owners,
            threshold=threshold,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
        )

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx(
            {
                "to": safe_creation_tx.safe_address,
                "value": safe_creation_tx.payment,
            },
            funder_account,
        )

        funder_balance = self.ethereum_client.get_balance(funder_account.address)
        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            funder_account,
            master_copy,
            safe_creation_tx.safe_setup_data,
            salt_nonce,
            safe_creation_tx.gas,
            safe_creation_tx.gas_price,
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt["status"], 1)

        # Funder balance must be bigger after a Safe deployment, as Safe deployment is a little overpriced
        self.assertGreater(
            self.ethereum_client.get_balance(funder_account.address), funder_balance
        )
        logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
            tx_receipt
        )
        log = logs[0]
        self.assertIsNone(tx_receipt["contractAddress"])
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
        )

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(
            deployed_safe_proxy_contract.functions.VERSION().call(), "1.0.0"
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getThreshold().call(), threshold
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getOwners().call(), owners
        )
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_with_payment_receiver(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        payment_receiver = Account.create().address
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price

        safe_creation_tx = SafeCreate2TxBuilder(
            w3=w3,
            master_copy_address=self.safe_contract.address,
            proxy_factory_address=self.proxy_factory_contract.address,
        ).build(
            owners=owners,
            threshold=threshold,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
            payment_receiver=payment_receiver,
        )

        self.assertEqual(safe_creation_tx.payment, safe_creation_tx.payment_ether)
        self.send_tx(
            {
                "to": safe_creation_tx.safe_address,
                "value": safe_creation_tx.payment,
            },
            funder_account,
        )

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            funder_account,
            self.safe_contract.address,
            safe_creation_tx.safe_setup_data,
            salt_nonce,
            gas=safe_creation_tx.gas,
            gas_price=safe_creation_tx.gas_price,
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt["status"], 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
            tx_receipt
        )
        log = logs[0]
        self.assertIsNone(tx_receipt["contractAddress"])
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
        )

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getThreshold().call(), threshold
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getOwners().call(), owners
        )
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

        self.assertEqual(
            self.ethereum_client.get_balance(payment_receiver), safe_creation_tx.payment
        )

    def test_safe_create2_tx_builder_with_fixed_cost(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        fixed_creation_cost = 123  # Wei

        safe_creation_tx = SafeCreate2TxBuilder(
            w3=w3,
            master_copy_address=self.safe_contract.address,
            proxy_factory_address=self.proxy_factory_contract.address,
        ).build(
            owners=owners,
            threshold=threshold,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
            fixed_creation_cost=fixed_creation_cost,
        )

        self.assertEqual(safe_creation_tx.payment, fixed_creation_cost)
        self.assertEqual(
            safe_creation_tx.payment_ether,
            safe_creation_tx.gas * safe_creation_tx.gas_price,
        )

        self.send_tx(
            {
                "to": safe_creation_tx.safe_address,
                "value": safe_creation_tx.payment,
            },
            funder_account,
        )

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            funder_account,
            self.safe_contract.address,
            safe_creation_tx.safe_setup_data,
            salt_nonce,
            gas=safe_creation_tx.gas,
            gas_price=safe_creation_tx.gas_price,
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt["status"], 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
            tx_receipt
        )
        log = logs[0]
        self.assertIsNone(tx_receipt["contractAddress"])
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
        )

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getThreshold().call(), threshold
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getOwners().call(), owners
        )
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_create2_tx_builder_with_token_payment(self):
        w3 = self.w3

        salt_nonce = generate_salt_nonce()
        erc20_deployer = Account.create()
        funder_account = self.ethereum_test_account
        owners = [Account.create().address for _ in range(4)]
        threshold = len(owners) - 1
        gas_price = self.gas_price
        token_amount = int(1e18)
        erc20_contract = self.deploy_example_erc20(token_amount, erc20_deployer.address)
        self.assertEqual(
            erc20_contract.functions.balanceOf(erc20_deployer.address).call(),
            token_amount,
        )

        # Send something to the erc20 deployer
        self.send_tx(
            {"to": erc20_deployer.address, "value": w3.to_wei(1, "ether")},
            funder_account,
        )

        safe_creation_tx = SafeCreate2TxBuilder(
            w3=w3,
            master_copy_address=self.safe_contract.address,
            proxy_factory_address=self.proxy_factory_contract.address,
        ).build(
            owners=owners,
            threshold=threshold,
            salt_nonce=salt_nonce,
            gas_price=gas_price,
            payment_token=erc20_contract.address,
        )

        self.assertEqual(safe_creation_tx.payment_token, erc20_contract.address)
        self.assertGreater(safe_creation_tx.payment, 0)
        self.assertEqual(
            safe_creation_tx.payment_ether,
            safe_creation_tx.gas * safe_creation_tx.gas_price,
        )

        self.send_tx(
            erc20_contract.functions.transfer(
                safe_creation_tx.safe_address, safe_creation_tx.payment
            ).build_transaction({"from": erc20_deployer.address}),
            erc20_deployer,
        )
        self.assertEqual(
            erc20_contract.functions.balanceOf(safe_creation_tx.safe_address).call(),
            safe_creation_tx.payment,
        )

        ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
            funder_account,
            self.safe_contract.address,
            safe_creation_tx.safe_setup_data,
            salt_nonce,
            gas=safe_creation_tx.gas,
            gas_price=safe_creation_tx.gas_price,
        )
        tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
        self.assertEqual(tx_receipt["status"], 1)
        logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
            tx_receipt
        )
        self.assertEqual(len(logs), 1)
        log = logs[0]
        self.assertIsNone(tx_receipt["contractAddress"])
        self.assertEqual(log["event"], "ProxyCreation")
        proxy_address = log["args"]["proxy"]
        self.assertEqual(proxy_address, safe_creation_tx.safe_address)
        self.assertEqual(
            ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
        )

        deployed_safe_proxy_contract = get_safe_contract(w3, proxy_address)
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getThreshold().call(), threshold
        )
        self.assertEqual(
            deployed_safe_proxy_contract.functions.getOwners().call(), owners
        )
        self.assertEqual(self.ethereum_client.get_balance(proxy_address), 0)

    def test_safe_gas_with_multiple_owners(self):
        logger.info(
            "Test Safe Proxy create2 gas with multiple owners".center(
                LOG_TITLE_WIDTH, "-"
            )
        )
        w3 = self.w3
        funder_account = self.ethereum_test_account
        number_of_accounts = 10
        for i in range(2, number_of_accounts):
            salt_nonce = generate_salt_nonce()
            owners = [Account.create().address for _ in range(i + 1)]
            threshold = len(owners) - 1
            gas_price = self.gas_price

            safe_creation_tx = SafeCreate2TxBuilder(
                w3=w3,
                master_copy_address=self.safe_contract.address,
                proxy_factory_address=self.proxy_factory_contract.address,
            ).build(
                owners=owners,
                threshold=threshold,
                salt_nonce=salt_nonce,
                gas_price=gas_price,
            )

            self.send_tx(
                {
                    "to": safe_creation_tx.safe_address,
                    "value": safe_creation_tx.payment,
                },
                funder_account,
            )

            ethereum_tx_sent = self.proxy_factory.deploy_proxy_contract_with_nonce(
                funder_account,
                self.safe_contract.address,
                safe_creation_tx.safe_setup_data,
                salt_nonce,
                gas=safe_creation_tx.gas,
                gas_price=safe_creation_tx.gas_price,
            )
            tx_receipt = w3.eth.wait_for_transaction_receipt(ethereum_tx_sent.tx_hash)
            self.assertEqual(tx_receipt["status"], 1)
            logs = self.proxy_factory_contract.events.ProxyCreation().process_receipt(
                tx_receipt
            )
            log = logs[0]
            self.assertIsNone(tx_receipt["contractAddress"])
            self.assertEqual(log["event"], "ProxyCreation")
            proxy_address = log["args"]["proxy"]
            self.assertEqual(proxy_address, safe_creation_tx.safe_address)
            self.assertEqual(
                ethereum_tx_sent.contract_address, safe_creation_tx.safe_address
            )

            logger.info(
                "Number of owners: %d - Gas estimated %d - Gas Used %d - Difference %d - Gas used per owner %d",
                len(owners),
                safe_creation_tx.gas,
                tx_receipt["gasUsed"],
                safe_creation_tx.gas - tx_receipt["gasUsed"],
                tx_receipt["gasUsed"] // len(owners),
            )



from time import time

from django.test import TestCase

import pytest
from eth_account import Account

from ...eth import EthereumNetwork
from ...eth.constants import NULL_ADDRESS
from .. import CowSwapAPI, Order, OrderKind


class TestCowSwapAPI(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.mainnet_cow_swap_api = CowSwapAPI(EthereumNetwork.MAINNET)
        cls.sepolia_cow_swap_api = CowSwapAPI(EthereumNetwork.SEPOLIA)
        cls.mainnet_gno_token_address = "0x6810e776880C02933D47DB1b9fc05908e5386b96"
        cls.sepolia_cow_token_address = "0x0625aFB445C3B6B7B929342a04A22599fd5dBB59"

    def test_api_is_available(self):
        random_owner = Account.create().address
        for ethereum_network in (
            EthereumNetwork.MAINNET,
            EthereumNetwork.SEPOLIA,
            EthereumNetwork.GNOSIS,
        ):
            with self.subTest(ethereum_network=ethereum_network):
                self.assertEqual(self.sepolia_cow_swap_api.get_orders(random_owner), [])

    def test_get_estimated_amount(self):
        cow_swap = CowSwapAPI(EthereumNetwork.MAINNET)
        response = cow_swap.get_estimated_amount(
            self.mainnet_gno_token_address,
            self.mainnet_gno_token_address,
            OrderKind.SELL,
            1,
        )
        self.assertDictEqual(
            response,
            {
                "errorType": "SameBuyAndSellToken",
                "description": "Buy token is the same as the sell token.",
            },
        )

        response = cow_swap.get_estimated_amount(
            "0x6820e776880c02933d47db1b9fc05908e5386b96",
            self.mainnet_gno_token_address,
            OrderKind.SELL,
            1,
        )
        self.assertIn("errorType", response)
        self.assertIn("description", response)

        response = cow_swap.get_estimated_amount(
            self.mainnet_gno_token_address,
            self.mainnet_cow_swap_api.weth_address,
            OrderKind.SELL,
            int(1e18),
        )
        amount = float(response["buyAmount"]) / response["sellAmount"]
        self.assertGreater(amount, 0)
        self.assertLess(amount, 1)

    def test_get_fee(self):
        order = Order(
            sellToken=self.mainnet_gno_token_address,
            buyToken=self.mainnet_gno_token_address,
            receiver=NULL_ADDRESS,
            sellAmount=int(1e18),
            buyAmount=1,
            validTo=int(time()) + 3600,
            appData={"version": "1.2.2", "metadata": {}},
            feeAmount=0,
            kind="sell",
            partiallyFillable=False,
            sellTokenBalance="erc20",
            buyTokenBalance="erc20",
        )
        from_address = Account.create().address
        self.assertEqual(
            self.mainnet_cow_swap_api.get_fee(order, from_address),
            {
                "errorType": "SameBuyAndSellToken",
                "description": "Buy token is the same as the sell token.",
            },
        )
        order.buyToken = self.mainnet_cow_swap_api.weth_address
        self.assertGreaterEqual(
            self.mainnet_cow_swap_api.get_fee(order, from_address), 0
        )

    def test_get_trades(self):
        mainnet_order_ui = "0x65F1206182C77A040ED41D507B59C622FA94AB5E71CCA567202CFF3909F3D5C4DBE338E45276630FD8237149DD47EE027AF26F9C619723D0"
        self.assertEqual(
            self.mainnet_cow_swap_api.get_trades(order_ui=mainnet_order_ui),
            [
                {
                    "blockNumber": 13643462,
                    "logIndex": 0,
                    "orderUid": "0x65f1206182c77a040ed41d507b59c622fa94ab5e71cca567202cff3909f3d5c4dbe338e45276630fd8237149dd47ee027af26f9c619723d0",
                    "buyAmount": "28361861093850079821",
                    "sellAmount": "113521821882",
                    "sellAmountBeforeFees": "113465370931",
                    "owner": "0xdbe338e45276630fd8237149dd47ee027af26f9c",
                    "buyToken": "0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2",
                    "sellToken": "0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48",
                    "txHash": "0x691d1a8ba39c036e841b6e2ed970f9068ac4a27b61955afb852f11019f2ff4d8",
                }
            ],
        )

    def test_place_order(self):
        order = Order(
            sellToken=self.mainnet_gno_token_address,
            buyToken=self.mainnet_gno_token_address,
            receiver=NULL_ADDRESS,
            sellAmount=1,
            buyAmount=1,
            validTo=int(time()) + 3600,
            appData={},
            feeAmount=0,
            kind="sell",
            partiallyFillable=False,
            sellTokenBalance="erc20",
            buyTokenBalance="erc20",
        )
        result = self.sepolia_cow_swap_api.place_order(order, Account().create().key)
        self.assertEqual(
            order.feeAmount, 0
        )  # Cannot estimate, as buy token is the same as the sell token
        self.assertEqual(
            result,
            {
                "description": "Buy token is the same as the sell token.",
                "errorType": "SameBuyAndSellToken",
            },
        )

        order.sellToken = self.sepolia_cow_swap_api.weth_address
        order.buyToken = self.sepolia_cow_token_address
        order_id = self.sepolia_cow_swap_api.place_order(order, Account().create().key)

        if type(order_id) is dict:
            pytest.xfail(order_id["errorType"])
        else:
            self.assertEqual(order_id[:2], "0x")
            self.assertEqual(len(order_id), 114)



from unittest import TestCase

from gnosis.eth.account_abstraction import UserOperation
from gnosis.eth.tests.mocks.mock_bundler import (
    safe_4337_chain_id_mock,
    safe_4337_module_address_mock,
    safe_4337_module_domain_separator_mock,
    safe_4337_safe_operation_hash_mock,
    safe_4337_user_operation_hash_mock,
    user_operation_mock,
)

from ...account_abstraction import SafeOperation
from ...account_abstraction.safe_operation import _domain_separator_cache


class TestSafeOperation(TestCase):
    def setUp(self):
        _domain_separator_cache.clear()

    def tearDown(self):
        _domain_separator_cache.clear()

    def test_safe_operation(self):
        safe_operation = SafeOperation.from_user_operation(
            UserOperation.from_bundler_response(
                safe_4337_user_operation_hash_mock, user_operation_mock["result"]
            )
        )

        self.assertDictEqual(_domain_separator_cache, {})

        self.assertEqual(
            safe_operation.get_domain_separator(
                safe_4337_chain_id_mock, safe_4337_module_address_mock
            ),
            safe_4337_module_domain_separator_mock,
        )
        self.assertDictEqual(
            _domain_separator_cache,
            {
                (
                    safe_4337_chain_id_mock,
                    safe_4337_module_address_mock,
                ): safe_4337_module_domain_separator_mock
            },
        )

        self.assertEqual(
            safe_operation.get_safe_operation_hash(
                safe_4337_chain_id_mock, safe_4337_module_address_mock
            ),
            safe_4337_safe_operation_hash_mock,
        )
        self.assertDictEqual(
            _domain_separator_cache,
            {
                (
                    safe_4337_chain_id_mock,
                    safe_4337_module_address_mock,
                ): safe_4337_module_domain_separator_mock
            },
        )



from eth_abi.exceptions import DecodingError
from web3.exceptions import Web3Exception

from .. import EthereumClient, EthereumNetwork
from .abis.superfluid_abis import super_token_abi
from .exceptions import CannotGetPriceFromOracle
from .oracles import PriceOracle


class SuperfluidOracle(PriceOracle):
    def __init__(self, ethereum_client: EthereumClient, price_oracle: PriceOracle):
        """
        :param ethereum_client:
        :param price_oracle: Price oracle to get the price for the components of Superfluid Tokens
        """
        self.ethereum_client = ethereum_client
        self.w3 = ethereum_client.w3
        self.price_oracle = price_oracle

    @classmethod
    def is_available(
        cls,
        ethereum_client: EthereumClient,
    ) -> bool:
        """
        :param ethereum_client:
        :return: `True` if Oracle is available for the EthereumClient provided, `False` otherwise
        """
        return ethereum_client.get_network() in (
            EthereumNetwork.POLYGON,
            EthereumNetwork.GNOSIS,
            EthereumNetwork.ARBITRUM_ONE,
            EthereumNetwork.OPTIMISM,
        )

    def get_price(self, token_address: str) -> float:
        try:
            underlying_token = (
                self.w3.eth.contract(token_address, abi=super_token_abi)
                .functions.getUnderlyingToken()
                .call()
            )
            return self.price_oracle.get_price(underlying_token)
        except (Web3Exception, DecodingError, ValueError):
            raise CannotGetPriceFromOracle(
                f"Cannot get price for {token_address}. It is not a wrapper Super Token"
            )



import django_filters

from .forms import EthereumAddressFieldForm, Keccak256FieldForm


class EthereumAddressFilter(django_filters.Filter):
    field_class = EthereumAddressFieldForm


class Keccak256Filter(django_filters.Filter):
    field_class = Keccak256FieldForm



from unittest import TestCase

from ...account_abstraction import UserOperation
from ..mocks.mock_bundler import (
    safe_4337_chain_id_mock,
    safe_4337_user_operation_hash_mock,
    user_operation_mock,
)


class TestUserOperation(TestCase):
    def test_calculate_user_operation_hash(self):
        user_operation_hash = safe_4337_user_operation_hash_mock.hex()
        user_operation = UserOperation.from_bundler_response(
            user_operation_hash, user_operation_mock["result"]
        )
        self.assertEqual(
            user_operation.calculate_user_operation_hash(safe_4337_chain_id_mock),
            safe_4337_user_operation_hash_mock,
        )



curve_address_provider_abi = [
    {
        "name": "NewAddressIdentifier",
        "inputs": [
            {"type": "uint256", "name": "id", "indexed": True},
            {"type": "address", "name": "addr", "indexed": False},
            {"type": "string", "name": "description", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "AddressModified",
        "inputs": [
            {"type": "uint256", "name": "id", "indexed": True},
            {"type": "address", "name": "new_address", "indexed": False},
            {"type": "uint256", "name": "version", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "CommitNewAdmin",
        "inputs": [
            {"type": "uint256", "name": "deadline", "indexed": True},
            {"type": "address", "name": "admin", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "NewAdmin",
        "inputs": [{"type": "address", "name": "admin", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "outputs": [],
        "inputs": [{"type": "address", "name": "_admin"}],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "name": "get_registry",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 1061,
    },
    {
        "name": "max_id",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 1258,
    },
    {
        "name": "get_address",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "uint256", "name": "_id"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 1308,
    },
    {
        "name": "add_new_id",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [
            {"type": "address", "name": "_address"},
            {"type": "string", "name": "_description"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 291275,
    },
    {
        "name": "set_address",
        "outputs": [{"type": "bool", "name": ""}],
        "inputs": [
            {"type": "uint256", "name": "_id"},
            {"type": "address", "name": "_address"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 182430,
    },
    {
        "name": "unset_address",
        "outputs": [{"type": "bool", "name": ""}],
        "inputs": [{"type": "uint256", "name": "_id"}],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 101348,
    },
    {
        "name": "commit_transfer_ownership",
        "outputs": [{"type": "bool", "name": ""}],
        "inputs": [{"type": "address", "name": "_new_admin"}],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 74048,
    },
    {
        "name": "apply_transfer_ownership",
        "outputs": [{"type": "bool", "name": ""}],
        "inputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 60125,
    },
    {
        "name": "revert_transfer_ownership",
        "outputs": [{"type": "bool", "name": ""}],
        "inputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 21400,
    },
    {
        "name": "admin",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 1331,
    },
    {
        "name": "transfer_ownership_deadline",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 1361,
    },
    {
        "name": "future_admin",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 1391,
    },
    {
        "name": "get_id_info",
        "outputs": [
            {"type": "address", "name": "addr"},
            {"type": "bool", "name": "is_active"},
            {"type": "uint256", "name": "version"},
            {"type": "uint256", "name": "last_modified"},
            {"type": "string", "name": "description"},
        ],
        "inputs": [{"type": "uint256", "name": "arg0"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 12168,
    },
]

curve_registry_abi = [
    {
        "name": "PoolAdded",
        "inputs": [
            {"type": "address", "name": "pool", "indexed": True},
            {"type": "bytes", "name": "rate_method_id", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "PoolRemoved",
        "inputs": [{"type": "address", "name": "pool", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_address_provider"},
            {"type": "address", "name": "_gauge_controller"},
        ],
        "stateMutability": "nonpayable",
        "type": "constructor",
    },
    {
        "name": "find_pool_for_coins",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [
            {"type": "address", "name": "_from"},
            {"type": "address", "name": "_to"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "name": "find_pool_for_coins",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [
            {"type": "address", "name": "_from"},
            {"type": "address", "name": "_to"},
            {"type": "uint256", "name": "i"},
        ],
        "stateMutability": "view",
        "type": "function",
    },
    {
        "name": "get_n_coins",
        "outputs": [{"type": "uint256[2]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 1704,
    },
    {
        "name": "get_coins",
        "outputs": [{"type": "address[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 12285,
    },
    {
        "name": "get_underlying_coins",
        "outputs": [{"type": "address[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 12347,
    },
    {
        "name": "get_decimals",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 8199,
    },
    {
        "name": "get_underlying_decimals",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 8261,
    },
    {
        "name": "get_rates",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 34780,
    },
    {
        "name": "get_gauges",
        "outputs": [
            {"type": "address[10]", "name": ""},
            {"type": "int128[10]", "name": ""},
        ],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 20310,
    },
    {
        "name": "get_balances",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 16818,
    },
    {
        "name": "get_underlying_balances",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 158953,
    },
    {
        "name": "get_virtual_price_from_lp_token",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [{"type": "address", "name": "_token"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 2080,
    },
    {
        "name": "get_A",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 1198,
    },
    {
        "name": "get_parameters",
        "outputs": [
            {"type": "uint256", "name": "A"},
            {"type": "uint256", "name": "future_A"},
            {"type": "uint256", "name": "fee"},
            {"type": "uint256", "name": "admin_fee"},
            {"type": "uint256", "name": "future_fee"},
            {"type": "uint256", "name": "future_admin_fee"},
            {"type": "address", "name": "future_owner"},
            {"type": "uint256", "name": "initial_A"},
            {"type": "uint256", "name": "initial_A_time"},
            {"type": "uint256", "name": "future_A_time"},
        ],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 6458,
    },
    {
        "name": "get_fees",
        "outputs": [{"type": "uint256[2]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 1603,
    },
    {
        "name": "get_admin_balances",
        "outputs": [{"type": "uint256[8]", "name": ""}],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 36719,
    },
    {
        "name": "get_coin_indices",
        "outputs": [
            {"type": "int128", "name": ""},
            {"type": "int128", "name": ""},
            {"type": "bool", "name": ""},
        ],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "address", "name": "_from"},
            {"type": "address", "name": "_to"},
        ],
        "stateMutability": "view",
        "type": "function",
        "gas": 27456,
    },
    {
        "name": "estimate_gas_used",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "address", "name": "_from"},
            {"type": "address", "name": "_to"},
        ],
        "stateMutability": "view",
        "type": "function",
        "gas": 32329,
    },
    {
        "name": "add_pool",
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "uint256", "name": "_n_coins"},
            {"type": "address", "name": "_lp_token"},
            {"type": "bytes32", "name": "_rate_method_id"},
            {"type": "uint256", "name": "_decimals"},
            {"type": "uint256", "name": "_underlying_decimals"},
            {"type": "bool", "name": "_has_initial_A"},
            {"type": "bool", "name": "_is_v1"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 10196577,
    },
    {
        "name": "add_pool_without_underlying",
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "uint256", "name": "_n_coins"},
            {"type": "address", "name": "_lp_token"},
            {"type": "bytes32", "name": "_rate_method_id"},
            {"type": "uint256", "name": "_decimals"},
            {"type": "uint256", "name": "_use_rates"},
            {"type": "bool", "name": "_has_initial_A"},
            {"type": "bool", "name": "_is_v1"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 5590664,
    },
    {
        "name": "add_metapool",
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "uint256", "name": "_n_coins"},
            {"type": "address", "name": "_lp_token"},
            {"type": "uint256", "name": "_decimals"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 10226976,
    },
    {
        "name": "remove_pool",
        "outputs": [],
        "inputs": [{"type": "address", "name": "_pool"}],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 779646579509,
    },
    {
        "name": "set_pool_gas_estimates",
        "outputs": [],
        "inputs": [
            {"type": "address[5]", "name": "_addr"},
            {"type": "uint256[2][5]", "name": "_amount"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 355578,
    },
    {
        "name": "set_coin_gas_estimates",
        "outputs": [],
        "inputs": [
            {"type": "address[10]", "name": "_addr"},
            {"type": "uint256[10]", "name": "_amount"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 357165,
    },
    {
        "name": "set_gas_estimate_contract",
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "address", "name": "_estimator"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 37747,
    },
    {
        "name": "set_liquidity_gauges",
        "outputs": [],
        "inputs": [
            {"type": "address", "name": "_pool"},
            {"type": "address[10]", "name": "_liquidity_gauges"},
        ],
        "stateMutability": "nonpayable",
        "type": "function",
        "gas": 365793,
    },
    {
        "name": "address_provider",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 2111,
    },
    {
        "name": "gauge_controller",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 2141,
    },
    {
        "name": "pool_list",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "uint256", "name": "arg0"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 2280,
    },
    {
        "name": "pool_count",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "stateMutability": "view",
        "type": "function",
        "gas": 2201,
    },
    {
        "name": "get_pool_from_lp_token",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "address", "name": "arg0"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 2446,
    },
    {
        "name": "get_lp_token",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "address", "name": "arg0"}],
        "stateMutability": "view",
        "type": "function",
        "gas": 2476,
    },
]

curve_pool_abi = [
    {
        "name": "TokenExchange",
        "inputs": [
            {"type": "address", "name": "buyer", "indexed": True},
            {"type": "int128", "name": "sold_id", "indexed": False},
            {"type": "uint256", "name": "tokens_sold", "indexed": False},
            {"type": "int128", "name": "bought_id", "indexed": False},
            {"type": "uint256", "name": "tokens_bought", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "TokenExchangeUnderlying",
        "inputs": [
            {"type": "address", "name": "buyer", "indexed": True},
            {"type": "int128", "name": "sold_id", "indexed": False},
            {"type": "uint256", "name": "tokens_sold", "indexed": False},
            {"type": "int128", "name": "bought_id", "indexed": False},
            {"type": "uint256", "name": "tokens_bought", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "AddLiquidity",
        "inputs": [
            {"type": "address", "name": "provider", "indexed": True},
            {"type": "uint256[4]", "name": "token_amounts", "indexed": False},
            {"type": "uint256[4]", "name": "fees", "indexed": False},
            {"type": "uint256", "name": "invariant", "indexed": False},
            {"type": "uint256", "name": "token_supply", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "RemoveLiquidity",
        "inputs": [
            {"type": "address", "name": "provider", "indexed": True},
            {"type": "uint256[4]", "name": "token_amounts", "indexed": False},
            {"type": "uint256[4]", "name": "fees", "indexed": False},
            {"type": "uint256", "name": "token_supply", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "RemoveLiquidityImbalance",
        "inputs": [
            {"type": "address", "name": "provider", "indexed": True},
            {"type": "uint256[4]", "name": "token_amounts", "indexed": False},
            {"type": "uint256[4]", "name": "fees", "indexed": False},
            {"type": "uint256", "name": "invariant", "indexed": False},
            {"type": "uint256", "name": "token_supply", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "CommitNewAdmin",
        "inputs": [
            {"type": "uint256", "name": "deadline", "indexed": True, "unit": "sec"},
            {"type": "address", "name": "admin", "indexed": True},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "NewAdmin",
        "inputs": [{"type": "address", "name": "admin", "indexed": True}],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "CommitNewParameters",
        "inputs": [
            {"type": "uint256", "name": "deadline", "indexed": True, "unit": "sec"},
            {"type": "uint256", "name": "A", "indexed": False},
            {"type": "uint256", "name": "fee", "indexed": False},
            {"type": "uint256", "name": "admin_fee", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "name": "NewParameters",
        "inputs": [
            {"type": "uint256", "name": "A", "indexed": False},
            {"type": "uint256", "name": "fee", "indexed": False},
            {"type": "uint256", "name": "admin_fee", "indexed": False},
        ],
        "anonymous": False,
        "type": "event",
    },
    {
        "outputs": [],
        "inputs": [
            {"type": "address[4]", "name": "_coins"},
            {"type": "address[4]", "name": "_underlying_coins"},
            {"type": "address", "name": "_pool_token"},
            {"type": "uint256", "name": "_A"},
            {"type": "uint256", "name": "_fee"},
        ],
        "constant": False,
        "payable": False,
        "type": "constructor",
    },
    {
        "name": "get_virtual_price",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 1570535,
    },
    {
        "name": "calc_token_amount",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [
            {"type": "uint256[4]", "name": "amounts"},
            {"type": "bool", "name": "deposit"},
        ],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 6103471,
    },
    {
        "name": "add_liquidity",
        "outputs": [],
        "inputs": [
            {"type": "uint256[4]", "name": "amounts"},
            {"type": "uint256", "name": "min_mint_amount"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 9331701,
    },
    {
        "name": "get_dy",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [
            {"type": "int128", "name": "i"},
            {"type": "int128", "name": "j"},
            {"type": "uint256", "name": "dx"},
        ],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 3489637,
    },
    {
        "name": "get_dy_underlying",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [
            {"type": "int128", "name": "i"},
            {"type": "int128", "name": "j"},
            {"type": "uint256", "name": "dx"},
        ],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 3489467,
    },
    {
        "name": "exchange",
        "outputs": [],
        "inputs": [
            {"type": "int128", "name": "i"},
            {"type": "int128", "name": "j"},
            {"type": "uint256", "name": "dx"},
            {"type": "uint256", "name": "min_dy"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 7034253,
    },
    {
        "name": "exchange_underlying",
        "outputs": [],
        "inputs": [
            {"type": "int128", "name": "i"},
            {"type": "int128", "name": "j"},
            {"type": "uint256", "name": "dx"},
            {"type": "uint256", "name": "min_dy"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 7050488,
    },
    {
        "name": "remove_liquidity",
        "outputs": [],
        "inputs": [
            {"type": "uint256", "name": "_amount"},
            {"type": "uint256[4]", "name": "min_amounts"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 241191,
    },
    {
        "name": "remove_liquidity_imbalance",
        "outputs": [],
        "inputs": [
            {"type": "uint256[4]", "name": "amounts"},
            {"type": "uint256", "name": "max_burn_amount"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 9330864,
    },
    {
        "name": "commit_new_parameters",
        "outputs": [],
        "inputs": [
            {"type": "uint256", "name": "amplification"},
            {"type": "uint256", "name": "new_fee"},
            {"type": "uint256", "name": "new_admin_fee"},
        ],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 146045,
    },
    {
        "name": "apply_new_parameters",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 133452,
    },
    {
        "name": "revert_new_parameters",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 21775,
    },
    {
        "name": "commit_transfer_ownership",
        "outputs": [],
        "inputs": [{"type": "address", "name": "_owner"}],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 74452,
    },
    {
        "name": "apply_transfer_ownership",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 60508,
    },
    {
        "name": "revert_transfer_ownership",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 21865,
    },
    {
        "name": "withdraw_admin_fees",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 23448,
    },
    {
        "name": "kill_me",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 37818,
    },
    {
        "name": "unkill_me",
        "outputs": [],
        "inputs": [],
        "constant": False,
        "payable": False,
        "type": "function",
        "gas": 21955,
    },
    {
        "name": "coins",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "int128", "name": "arg0"}],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2130,
    },
    {
        "name": "underlying_coins",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [{"type": "int128", "name": "arg0"}],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2160,
    },
    {
        "name": "balances",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [{"type": "int128", "name": "arg0"}],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2190,
    },
    {
        "name": "A",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2021,
    },
    {
        "name": "fee",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2051,
    },
    {
        "name": "admin_fee",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2081,
    },
    {
        "name": "owner",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2111,
    },
    {
        "name": "admin_actions_deadline",
        "outputs": [{"type": "uint256", "unit": "sec", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2141,
    },
    {
        "name": "transfer_ownership_deadline",
        "outputs": [{"type": "uint256", "unit": "sec", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2171,
    },
    {
        "name": "future_A",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2201,
    },
    {
        "name": "future_fee",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2231,
    },
    {
        "name": "future_admin_fee",
        "outputs": [{"type": "uint256", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2261,
    },
    {
        "name": "future_owner",
        "outputs": [{"type": "address", "name": ""}],
        "inputs": [],
        "constant": True,
        "payable": False,
        "type": "function",
        "gas": 2291,
    },
]



from unittest import mock
from unittest.mock import MagicMock

from django.test import TestCase

import pytest
from eth_account import Account
from eth_typing import ChecksumAddress

from .. import EthereumClient
from ..oracles import (
    AaveOracle,
    BalancerOracle,
    CannotGetPriceFromOracle,
    CreamOracle,
    CurveOracle,
    EnzymeOracle,
    MooniswapOracle,
    PoolTogetherOracle,
    SushiswapOracle,
    UniswapOracle,
    UniswapV2Oracle,
    YearnOracle,
    ZerionComposedOracle,
)
from ..oracles.utils import get_decimals as oracles_get_decimals
from .ethereum_test_case import EthereumTestCaseMixin
from .utils import just_test_if_mainnet_node

gno_token_mainnet_address = "0x6810e776880C02933D47DB1b9fc05908e5386b96"
weth_token_mainnet_address = "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2"
wbtc_token_mainnet_address = "0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599"

dai_token_mainnet_address = "0x6B175474E89094C44Da98b954EedeAC495271d0F"
usdc_token_mainnet_address = "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"
usdt_token_mainnet_address = "0xdAC17F958D2ee523a2206206994597C13D831ec7"


class TestOracles(EthereumTestCaseMixin, TestCase):
    def test_uniswap_oracle(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(UniswapOracle.is_available(ethereum_client))
        uniswap_oracle = UniswapOracle(ethereum_client)
        token_address = dai_token_mainnet_address
        price = uniswap_oracle.get_price(token_address)
        self.assertEqual(uniswap_oracle.get_uniswap_exchange.cache_info().hits, 0)
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, "Not enough liquidity"):
            token_address = gno_token_mainnet_address
            uniswap_oracle.get_price(token_address)

        # Check batching is working
        uniswap_exchange = uniswap_oracle.get_uniswap_exchange(token_address)
        self.assertEqual(uniswap_oracle.get_uniswap_exchange.cache_info().hits, 1)
        self.assertEqual(
            uniswap_oracle._get_balances_using_batching(
                uniswap_exchange, token_address
            ),
            uniswap_oracle._get_balances_without_batching(
                uniswap_exchange, token_address
            ),
        )

    def test_uniswap_oracle_not_deployed(self):
        uniswap_oracle = UniswapOracle(self.ethereum_client, Account.create().address)
        random_token_address = Account.create().address
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle, "Non existing uniswap exchange"
        ):
            uniswap_oracle.get_price(random_token_address)


class TestUniswapV2Oracle(EthereumTestCaseMixin, TestCase):
    @mock.patch.object(
        UniswapV2Oracle,
        "factory_address",
        return_value="0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f",
        new_callable=mock.PropertyMock,
    )
    def test_calculate_pair_address(self, factory_address_mock: MagicMock):
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)
        expected_address = "0x3e8468f66d30Fc99F745481d4B383f89861702C6"

        self.assertEqual(
            uniswap_v2_oracle.calculate_pair_address(
                gno_token_mainnet_address, weth_token_mainnet_address
            ),
            expected_address,
        )
        self.assertEqual(
            uniswap_v2_oracle.calculate_pair_address(
                weth_token_mainnet_address, gno_token_mainnet_address
            ),
            expected_address,
        )

    def test_get_price(self):
        oracles_get_decimals.cache_clear()
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(UniswapV2Oracle.is_available(ethereum_client))
        uniswap_v2_oracle = UniswapV2Oracle(ethereum_client)

        self.assertEqual(oracles_get_decimals.cache_info().currsize, 0)

        price = uniswap_v2_oracle.get_price(
            gno_token_mainnet_address, weth_token_mainnet_address
        )
        self.assertLess(price, 1)
        self.assertGreater(price, 0)

        self.assertEqual(oracles_get_decimals.cache_info().currsize, 2)

        # Test with 2 stablecoins
        price = uniswap_v2_oracle.get_price(
            dai_token_mainnet_address, usdt_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)
        self.assertEqual(oracles_get_decimals.cache_info().currsize, 4)
        self.assertEqual(oracles_get_decimals.cache_info().hits, 0)

        price = uniswap_v2_oracle.get_price(
            usdt_token_mainnet_address, dai_token_mainnet_address
        )
        self.assertAlmostEqual(price, 1.0, delta=0.5)
        self.assertEqual(oracles_get_decimals.cache_info().currsize, 4)
        self.assertEqual(oracles_get_decimals.cache_info().hits, 2)
        oracles_get_decimals.cache_clear()

    def test_get_price_contract_not_deployed(self):
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)
        random_token_address = Account.create().address
        with self.assertRaisesMessage(
            CannotGetPriceFromOracle,
            f"Cannot get uniswap v2 price for pair token_1={random_token_address}",
        ):
            uniswap_v2_oracle.get_price(random_token_address)

    @mock.patch("gnosis.eth.oracles.oracles.get_decimals", autospec=True)
    @mock.patch.object(
        UniswapV2Oracle,
        "factory_address",
        return_value="0x5C69bEe701ef814a2B6a3EDD4B1652CB9cc5aA6f",
        new_callable=mock.PropertyMock,
    )
    @mock.patch.object(
        UniswapV2Oracle, "get_reserves", return_value=(int(1e20), 600), autospec=True
    )
    def test_get_price_liquidity(
        self,
        get_reserves_mock: MagicMock,
        factory_address_mock: MagicMock,
        get_decimals_mock: MagicMock,
    ):
        token_1, token_2 = (
            "0xA14F6F8867DB84a45BCD148bfaf4e4f54B4B9b12",
            "0xC426A8F4C79EF274Ed93faC9e1A09bFC5659B06B",
        )

        def get_decimals_mock_fn(
            token_address: ChecksumAddress, ethereum_client: EthereumClient
        ) -> int:
            if token_address == token_1:
                return 18
            else:
                return 3

        get_decimals_mock.side_effect = get_decimals_mock_fn
        uniswap_v2_oracle = UniswapV2Oracle(self.ethereum_client)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, "Not enough liquidity"):
            uniswap_v2_oracle.get_price(token_1, token_2)

        get_reserves_mock.return_value = (int(1e20), 6000)
        self.assertEqual(uniswap_v2_oracle.get_price(token_1, token_2), 0.06)

        get_reserves_mock.return_value = reversed(get_reserves_mock.return_value)
        with self.assertRaisesMessage(CannotGetPriceFromOracle, "Not enough liquidity"):
            self.assertEqual(
                uniswap_v2_oracle.get_price(token_2, token_1), 0.06
            )  # Reserves were inverted

    def test_get_pool_token_price(self):
        dai_eth_pool_address = "0xA478c2975Ab1Ea89e8196811F51A7B7Ade33eB11"
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        uniswap_v2_oracle = UniswapV2Oracle(ethereum_client)

        price = uniswap_v2_oracle.get_pool_token_price(dai_eth_pool_address)
        self.assertGreater(price, 0.0)


class TestAaveOracle(EthereumTestCaseMixin, TestCase):
    def test_get_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(AaveOracle.is_available(ethereum_client))
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        aave_oracle = AaveOracle(ethereum_client, uniswap_oracle)

        aweth_address = "0x030bA81f1c18d280636F32af80b9AAd02Cf0854e"
        price = aave_oracle.get_price(aweth_address)
        self.assertGreater(price, 0.0)

        stacked_aave_address = "0x4da27a545c0c5B758a6BA100e3a049001de870f5"
        price = aave_oracle.get_price(stacked_aave_address)
        self.assertGreater(price, 0.0)

        error_message = "It is not an Aaave atoken"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            aave_oracle.get_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            aave_oracle.get_price(Account.create().address)


class TestCreamOracle(EthereumTestCaseMixin, TestCase):
    def test_get_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(CreamOracle.is_available(ethereum_client))
        sushi_oracle = SushiswapOracle(ethereum_client)
        cream_oracle = CreamOracle(ethereum_client, sushi_oracle)

        cyusdc_address = "0x76Eb2FE28b36B3ee97F3Adae0C69606eeDB2A37c"
        price = cream_oracle.get_price(cyusdc_address)
        self.assertGreater(price, 0.0)

        cydai_address = "0x8e595470Ed749b85C6F7669de83EAe304C2ec68F"
        price = cream_oracle.get_price(cydai_address)
        self.assertGreater(price, 0.0)

        error_message = "It is not a Cream cToken"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            cream_oracle.get_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            cream_oracle.get_price(Account.create().address)


class TestCurveOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_tokens(self):
        curve_token_address = (
            "0xC25a3A3b969415c80451098fa907EC722572917F"  # Curve.fi DAI/USDC/USDT/sUSD
        )
        local_curve_oracle = CurveOracle(self.ethereum_client)
        error_message = "Cannot find Zerion adapter"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            local_curve_oracle.get_underlying_tokens(curve_token_address)

        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(CurveOracle.is_available(ethereum_client))
        curve_oracle = CurveOracle(ethereum_client)

        # Curve.fi ETH/stETH (steCRV) is working with the updated adapter
        self.assertEqual(
            len(
                curve_oracle.get_underlying_tokens(
                    "0x06325440D014e39736583c165C2963BA99fAf14E"
                )
            ),
            2,
        )

        underlying_tokens = curve_oracle.get_underlying_tokens(curve_token_address)
        self.assertEqual(len(underlying_tokens), 4)

        self.assertCountEqual(
            [underlying_token.address for underlying_token in underlying_tokens],
            [
                "0x6B175474E89094C44Da98b954EedeAC495271d0F",
                "0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48",
                "0xdAC17F958D2ee523a2206206994597C13D831ec7",
                "0x57Ab1ec28D129707052df4dF418D58a2D46d5f51",
            ],
        )
        for underlying_token in underlying_tokens:
            self.assertTrue(0.0 < underlying_token.quantity < 1.0)

        error_message = "It is not a Zerion supported pool token"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            curve_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            curve_oracle.get_underlying_tokens(Account.create().address)

    def test_get_underlying_tokens_gauges(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)
        curve_oracle = CurveOracle(ethereum_client)

        # 3crv gauge deposit: dai, usdc, usdt
        gauge_deposit_address = "0xF5194c3325202F456c95c1Cf0cA36f8475C1949F"
        gauge_lp_token_address = "0x5282a4eF67D9C33135340fB3289cc1711c13638C"
        gauge_underlying_tokens = curve_oracle.get_underlying_tokens(
            gauge_deposit_address
        )
        lp_token_underlying_tokens = curve_oracle.get_underlying_tokens(
            gauge_lp_token_address
        )

        self.assertEqual(gauge_underlying_tokens, lp_token_underlying_tokens)


class TestZerionComposedOracle(EthereumTestCaseMixin, TestCase):
    def test_zerion_composed_oracle(self):
        with self.assertRaisesMessage(ValueError, "Expected a Zerion adapter address"):
            ZerionComposedOracle(self.ethereum_client)


class TestPoolTogetherOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_token(self):
        pool_together_token_address = (
            "0xD81b1A8B1AD00Baa2D6609E0BAE28A38713872f7"  # v3 USDC Ticket
        )
        local_pool_together_oracle = CurveOracle(self.ethereum_client)
        error_message = "Cannot find Zerion adapter"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            local_pool_together_oracle.get_underlying_tokens(
                pool_together_token_address
            )

        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(PoolTogetherOracle.is_available(ethereum_client))
        pool_together_oracle = PoolTogetherOracle(ethereum_client)

        underlying_tokens = pool_together_oracle.get_underlying_tokens(
            pool_together_token_address
        )
        self.assertEqual(len(underlying_tokens), 1)
        self.assertEqual(
            [underlying_token.address for underlying_token in underlying_tokens],
            ["0xA0b86991c6218b36c1d19D4a2e9Eb0cE3606eB48"],
        )
        self.assertEqual(underlying_tokens[0].quantity, 1.0)

        error_message = "It is not a Zerion supported pool token"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            pool_together_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            pool_together_oracle.get_underlying_tokens(Account.create().address)


class TestYearnOracle(EthereumTestCaseMixin, TestCase):
    def test_get_underlying_tokens(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(YearnOracle.is_available(ethereum_client))
        yearn_oracle = YearnOracle(ethereum_client)
        yearn_token_address = "0x5533ed0a3b83F70c3c4a1f69Ef5546D3D4713E44"  # Yearn Curve.fi DAI/USDC/USDT/sUSD
        yearn_underlying_token_address = (
            "0xC25a3A3b969415c80451098fa907EC722572917F"  # Curve.fi DAI/USDC/USDT/sUSD
        )
        iearn_token_address = "0x16de59092dAE5CcF4A1E6439D611fd0653f0Bd01"  # iearn DAI
        iearn_underlying_token_address = (
            "0x6B175474E89094C44Da98b954EedeAC495271d0F"  # DAI
        )
        yvault_token_address = (
            "0xdCD90C7f6324cfa40d7169ef80b12031770B4325"  # steCRV yVault
        )
        yvault_underlying_token_address = (
            "0x06325440D014e39736583c165C2963BA99fAf14E"  # steCRV
        )

        underlying_tokens = yearn_oracle.get_underlying_tokens(yearn_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.0)
        self.assertEqual(underlying_token.address, yearn_underlying_token_address)

        underlying_tokens = yearn_oracle.get_underlying_tokens(iearn_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.0)
        self.assertEqual(underlying_token.address, iearn_underlying_token_address)

        underlying_tokens = yearn_oracle.get_underlying_tokens(yvault_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.0)
        self.assertEqual(underlying_token.address, yvault_underlying_token_address)

        # Test yToken
        y_token = "0x30FCf7c6cDfC46eC237783D94Fc78553E79d4E9C"
        yearn_underlying_token = "0x3a664Ab939FD8482048609f652f9a0B0677337B9"
        underlying_tokens = yearn_oracle.get_underlying_tokens(y_token)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertAlmostEqual(underlying_token.quantity, 1.0, delta=0.5)
        self.assertEqual(underlying_token.address, yearn_underlying_token)

        error_message = "It is not a Yearn yToken/yVault"
        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            yearn_oracle.get_underlying_tokens(gno_token_mainnet_address)

        with self.assertRaisesMessage(CannotGetPriceFromOracle, error_message):
            yearn_oracle.get_underlying_tokens(Account.create().address)


class TestBalancerOracle(EthereumTestCaseMixin, TestCase):
    @pytest.mark.xfail(reason="Could fail due to unbalanced pool")
    def test_get_pool_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(BalancerOracle.is_available(ethereum_client))
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        balancer_oracle = BalancerOracle(ethereum_client, uniswap_oracle)
        balancer_token_address = (
            "0x59A19D8c652FA0284f44113D0ff9aBa70bd46fB4"  # Balancer 80% BAL + 20% WETH
        )

        price = balancer_oracle.get_pool_token_price(balancer_token_address)
        self.assertAlmostEqual(price, 1.0, delta=0.9)

        with self.assertRaisesMessage(
            CannotGetPriceFromOracle, "It is not a balancer pool token"
        ):
            balancer_oracle.get_pool_token_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(
            CannotGetPriceFromOracle, "It is not a balancer pool token"
        ):
            balancer_oracle.get_pool_token_price(Account.create().address)


class TestMooniswapOracle(EthereumTestCaseMixin, TestCase):
    def test_get_pool_token_price(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(MooniswapOracle.is_available(ethereum_client))
        uniswap_oracle = UniswapV2Oracle(ethereum_client)
        mooniswap_oracle = MooniswapOracle(ethereum_client, uniswap_oracle)
        mooniswap_pool_address = "0x6a11F3E5a01D129e566d783A7b6E8862bFD66CcA"  # 1inch Liquidity Pool (ETH-WBTC)
        other_version_mooniswap_pool_address = (
            "0x8a2f2F8637deb4Ee7C9400A240d27E3A32147dA4"  # Mooniswap V1 (OWL-GNO)
        )

        price = mooniswap_oracle.get_pool_token_price(mooniswap_pool_address)
        self.assertGreater(price, 0)

        price = mooniswap_oracle.get_pool_token_price(
            other_version_mooniswap_pool_address
        )
        self.assertGreater(price, 0)

        with self.assertRaisesMessage(
            CannotGetPriceFromOracle, "It is not a mooniswap pool token"
        ):
            mooniswap_oracle.get_pool_token_price(gno_token_mainnet_address)

        with self.assertRaisesMessage(
            CannotGetPriceFromOracle, "It is not a mooniswap pool token"
        ):
            mooniswap_oracle.get_pool_token_price(Account.create().address)


class TestEnzymeOracle(EthereumTestCaseMixin, TestCase):
    @pytest.mark.xfail(reason="Flaky because of external dependency on Enzyme")
    def test_get_underlying_tokens(self):
        mainnet_node = just_test_if_mainnet_node()
        ethereum_client = EthereumClient(mainnet_node)

        self.assertTrue(EnzymeOracle.is_available(ethereum_client))
        enzyme_oracle = EnzymeOracle(ethereum_client)
        mln_vault_token_address = "0x45c45799Bcf6C7Eb2Df0DA1240BE04cE1D18CC69"
        mln_vault_underlying_token = "0xec67005c4E498Ec7f55E092bd1d35cbC47C91892"
        usf_fund_token_address = "0x86FB84E92c1EEDc245987D28a42E123202bd6701"
        usf_fund_underlying_tokens = [
            "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
            "0x182B723a58739a9c974cFDB385ceaDb237453c28",
            "0x5A98FcBEA516Cf06857215779Fd812CA3beF1B32",
            "0xD533a949740bb3306d119CC777fa900bA034cd52",
        ]

        underlying_tokens = enzyme_oracle.get_underlying_tokens(mln_vault_token_address)
        self.assertEqual(len(underlying_tokens), 1)
        underlying_token = underlying_tokens[0]
        self.assertEqual(underlying_token.quantity, 1.0)
        self.assertEqual(underlying_token.address, mln_vault_underlying_token)

        underlying_tokens = enzyme_oracle.get_underlying_tokens(usf_fund_token_address)
        self.assertEqual(len(underlying_tokens), 2)
        for underlying_token in underlying_tokens:
            self.assertIn(underlying_token.address, usf_fund_underlying_tokens)
            self.assertAlmostEqual(underlying_token.quantity, 0.5, delta=0.5)



# flake8: noqa F401
from .ethereum_client import (
    EthereumClient,
    EthereumClientProvider,
    EthereumTxSent,
    FromAddressNotFound,
    GasLimitExceeded,
    InsufficientFunds,
    InvalidNonce,
    NonceTooHigh,
    NonceTooLow,
    ReplacementTransactionUnderpriced,
    SenderAccountNotFoundInNode,
    TransactionAlreadyImported,
    TransactionQueueLimitReached,
    TxSpeed,
    UnknownAccount,
)
from .ethereum_network import EthereumNetwork, EthereumNetworkNotSupported
from .exceptions import InvalidERC20Info, InvalidERC721Info

__all__ = [
    "EthereumClient",
    "EthereumClientProvider",
    "EthereumTxSent",
    "FromAddressNotFound",
    "GasLimitExceeded",
    "InsufficientFunds",
    "InvalidNonce",
    "NonceTooHigh",
    "NonceTooLow",
    "ReplacementTransactionUnderpriced",
    "SenderAccountNotFoundInNode",
    "TransactionAlreadyImported",
    "TransactionQueueLimitReached",
    "TxSpeed",
    "UnknownAccount",
    "EthereumNetwork",
    "EthereumNetworkNotSupported",
    "InvalidERC20Info",
    "InvalidERC721Info",
]
